import datetime as dt

from django import forms
from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.utils.html import mark_safe
from django.utils.text import capfirst
from django.utils.timezone import localtime
from django.utils.translation import gettext, gettext_lazy as _, override

from workbench.accounts.features import FEATURES
from workbench.accounts.models import User
from workbench.circles.models import Circle, Role
from workbench.contacts.models import Organization
from workbench.expenses.models import ExchangeRates
from workbench.logbook.models import Break, LoggedCost, LoggedHours
from workbench.offers.models import Offer
from workbench.projects.models import Campaign, Project, Service
from workbench.timer.models import Timestamp
from workbench.tools.forms import (
    Autocomplete,
    DateInput,
    Form,
    ModelForm,
    Textarea,
    add_prefix,
    querystring,
)
from workbench.tools.validation import (
    in_days,
    is_title_specific,
    logbook_lock,
    raise_if_errors,
)
from workbench.tools.xlsx import WorkbenchXLSXDocument


class DetectedTimestampForm(forms.Form):
    detected_ends_at = forms.CharField()

    def clean_detected_ends_at(self):
        value = parse_datetime(self.cleaned_data.get("detected_ends_at"))
        if not value:
            raise forms.ValidationError("Invalid value '%s'" % value)
        return value

    def build_if_valid(self, **kwargs):
        if self.is_valid():
            kwargs.setdefault("notes", gettext("<detected>"))
            return Timestamp(
                type=Timestamp.STOP,
                created_at=self.cleaned_data["detected_ends_at"],
                **kwargs,
            )


class LoggedHoursSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    rendered_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    project = forms.ModelChoiceField(
        queryset=Project.objects.all(),
        required=False,
        widget=Autocomplete(model=Project),
        label="",
    )
    organization = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    date_from = forms.DateField(widget=DateInput, required=False, label="")
    date_until = forms.DateField(
        widget=DateInput, required=False, label=mark_safe("&ndash;&nbsp;")
    )
    service = forms.ModelChoiceField(
        queryset=Service.objects.all(),
        required=False,
        widget=forms.HiddenInput,
        label="",
    )
    campaign = forms.ModelChoiceField(
        queryset=Campaign.objects.all(),
        required=False,
        widget=forms.HiddenInput,
        label="",
    )
    offer = forms.IntegerField(required=False, widget=forms.HiddenInput, label="")
    circle = forms.IntegerField(required=False, widget=forms.HiddenInput, label="")
    role = forms.ModelChoiceField(
        queryset=Role.objects.all(), required=False, widget=forms.HiddenInput, label=""
    )
    category = forms.CharField(required=False, widget=forms.HiddenInput, label="")
    not_archived = forms.BooleanField(
        required=False, widget=forms.HiddenInput, label=""
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["rendered_by"].choices = User.objects.choices(
            collapse_inactive=False, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("rendered_by") == -1:
            queryset = queryset.filter(rendered_by=self.request.user)
        elif data.get("rendered_by"):
            queryset = queryset.filter(rendered_by=data.get("rendered_by"))
        if data.get("project"):
            queryset = queryset.filter(service__project=data.get("project"))
        if data.get("organization"):
            queryset = queryset.filter(
                service__project__customer=data.get("organization")
            )
        if data.get("date_from"):
            queryset = queryset.filter(rendered_on__gte=data.get("date_from"))
        if data.get("date_until"):
            queryset = queryset.filter(rendered_on__lte=data.get("date_until"))

        # "hidden" filters
        self.hidden_filters = []
        if data.get("service"):
            service = data.get("service")
            self.hidden_filters.append(
                (
                    f"{capfirst(service._meta.verbose_name)}: {service}",
                    querystring(self.request.GET, service=""),
                ),
            )
            queryset = queryset.filter(service=service)
        if data.get("campaign"):
            campaign = data.get("campaign")
            self.hidden_filters.append(
                (
                    f"{capfirst(campaign._meta.verbose_name)}: {campaign}",
                    querystring(self.request.GET, campaign=""),
                ),
            )
            queryset = queryset.filter(service__project__campaign=campaign)
        if data.get("offer") == 0:
            self.hidden_filters.append(
                (_("No offer"), querystring(self.request.GET, offer=""))
            )
            queryset = queryset.filter(service__offer__isnull=True)
        elif data.get("offer"):
            offer = Offer.objects.filter(pk=data.get("offer")).first()
            self.hidden_filters.append(
                (
                    f"{capfirst(Offer._meta.verbose_name)}: {offer}",
                    querystring(self.request.GET, offer=""),
                ),
            )
            queryset = queryset.filter(service__offer=offer)
        if data.get("circle") == 0:
            queryset = queryset.filter(service__role__isnull=True)
            self.hidden_filters.append(
                (_("No circle"), querystring(self.request.GET, circle=""))
            )
        elif data.get("circle"):
            circle = Circle.objects.filter(pk=data.get("circle")).first()
            self.hidden_filters.append(
                (
                    f"{capfirst(Circle._meta.verbose_name)}: {circle}",
                    querystring(self.request.GET, circle=""),
                ),
            )
            queryset = queryset.filter(service__role__circle=circle)
        if data.get("role"):
            role = data.get("role")
            self.hidden_filters.append(
                (
                    f"{capfirst(role._meta.verbose_name)}: {role}",
                    querystring(self.request.GET, role=""),
                ),
            )
            queryset = queryset.filter(service__role=role)
        if data.get("category") == "none":
            self.hidden_filters.append(
                (
                    f"{capfirst(_('category'))}: {_('Undefined')}",
                    querystring(self.request.GET, category=""),
                ),
            )
            queryset = queryset.filter(
                Q(service__role__isnull=True) | Q(service__role__work_category="")
            )

        elif category := data.get("category"):
            self.hidden_filters.append(
                (
                    f"{capfirst(_('category'))}: {category}",
                    querystring(self.request.GET, category=""),
                ),
            )
            queryset = queryset.filter(service__role__work_category=category)
        if data.get("not_archived"):
            self.hidden_filters.append(
                (_("Not archived"), querystring(self.request.GET, not_archived=""))
            )
            queryset = queryset.filter(archived_at__isnull=True)

        return queryset.select_related("service__project__owned_by", "rendered_by")

    def response(self, request, queryset):
        if (
            request.GET.get("export") == "xlsx"
            and request.user.features[FEATURES.CONTROLLING]
        ):
            xlsx = WorkbenchXLSXDocument()
            xlsx.logged_hours(queryset)
            return xlsx.to_response("hours.xlsx")


class LoggedCostSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    rendered_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    project = forms.ModelChoiceField(
        queryset=Project.objects.all(),
        required=False,
        widget=Autocomplete(model=Project),
        label="",
    )
    organization = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    expenses = forms.BooleanField(required=False, label=_("expenses"))
    date_from = forms.DateField(widget=DateInput, required=False, label="")
    date_until = forms.DateField(
        widget=DateInput, required=False, label=mark_safe("&ndash;&nbsp;")
    )
    service = forms.ModelChoiceField(
        queryset=Service.objects.all(),
        required=False,
        widget=forms.HiddenInput,
        label="",
    )
    campaign = forms.ModelChoiceField(
        queryset=Campaign.objects.all(),
        required=False,
        widget=forms.HiddenInput,
        label="",
    )
    offer = forms.IntegerField(required=False, widget=forms.HiddenInput, label="")
    not_archived = forms.BooleanField(
        required=False, widget=forms.HiddenInput, label=""
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["rendered_by"].choices = User.objects.choices(
            collapse_inactive=False, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("rendered_by") == -1:
            queryset = queryset.filter(rendered_by=self.request.user)
        elif data.get("rendered_by"):
            queryset = queryset.filter(rendered_by=data.get("rendered_by"))
        if data.get("project"):
            queryset = queryset.filter(service__project=data.get("project"))
        if data.get("organization"):
            queryset = queryset.filter(
                service__project__customer=data.get("organization")
            )
        if data.get("expenses"):
            queryset = queryset.filter(are_expenses=True)
        if data.get("date_from"):
            queryset = queryset.filter(rendered_on__gte=data.get("date_from"))
        if data.get("date_until"):
            queryset = queryset.filter(rendered_on__lte=data.get("date_until"))

        # "hidden" filters
        self.hidden_filters = []
        if data.get("service"):
            service = data.get("service")
            self.hidden_filters.append(
                (
                    f"{capfirst(service._meta.verbose_name)}: {service}",
                    querystring(self.request.GET, service=""),
                ),
            )
            queryset = queryset.filter(service=service)
        if data.get("campaign"):
            campaign = data.get("campaign")
            self.hidden_filters.append(
                (
                    f"{capfirst(campaign._meta.verbose_name)}: {campaign}",
                    querystring(self.request.GET, campaign=""),
                ),
            )
            queryset = queryset.filter(service__project__campaign=campaign)
        if data.get("offer") == 0:
            self.hidden_filters.append(
                (_("No offer"), querystring(self.request.GET, offer=""))
            )
            queryset = queryset.filter(service__offer__isnull=True)
        elif data.get("offer"):
            offer = Offer.objects.filter(pk=data.get("offer")).first()
            self.hidden_filters.append(
                (
                    f"{capfirst(Offer._meta.verbose_name)}: {offer}",
                    querystring(self.request.GET, offer=""),
                ),
            )
            queryset = queryset.filter(service__offer=offer)
        if data.get("not_archived"):
            self.hidden_filters.append(
                (_("Not archived"), querystring(self.request.GET, not_archived=""))
            )
            queryset = queryset.filter(archived_at__isnull=True)

        return queryset.select_related("service__project__owned_by", "rendered_by")

    def response(self, request, queryset):
        if (
            request.GET.get("export") == "xlsx"
            and request.user.features[FEATURES.CONTROLLING]
        ):
            xlsx = WorkbenchXLSXDocument()
            xlsx.logged_costs(queryset)
            return xlsx.to_response("costs.xlsx")


@add_prefix("modal")
class LoggedHoursForm(ModelForm):
    user_fields = default_to_current_user = ("rendered_by",)

    service_title = forms.CharField(label=_("title"), required=False, max_length=200)
    service_description = forms.CharField(
        label=_("description"), required=False, widget=Textarea({"rows": 2})
    )
    service_role = forms.ModelChoiceField(
        queryset=Role.objects.all(), label=_("role"), required=False
    )

    class Meta:
        model = LoggedHours
        fields = ("rendered_by", "rendered_on", "service", "hours", "description")
        widgets = {"description": Textarea({"rows": 2})}

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        if self.project:
            initial = kwargs.setdefault("initial", {})
            request = kwargs["request"]

            if pk := request.GET.get("copy"):
                try:
                    hours = LoggedHours.objects.get(pk=pk)
                except (LoggedHours.DoesNotExist, TypeError, ValueError):
                    pass
                else:
                    initial.update(
                        {
                            "service": hours.service_id,
                            "rendered_on": hours.rendered_on.isoformat(),
                            "hours": hours.hours,
                            "description": hours.description,
                        }
                    )

            for field in ["description", "hours", "rendered_on", "service"]:
                if value := request.GET.get(field):
                    initial[field] = value

            if not initial.get("hours") and request.user.hours_since_latest <= 12:
                initial.setdefault("hours", request.user.hours_since_latest)

            if not initial.get("service"):
                latest_on_project = (
                    LoggedHours.objects.filter(
                        rendered_by=request.user, service__project=self.project
                    )
                    .order_by("-created_at")
                    .first()
                )
                if latest_on_project:
                    initial.setdefault("service", latest_on_project.service_id)

        else:
            self.project = kwargs["instance"].service.project

        super().__init__(*args, **kwargs)
        self.fields["service"].choices = self.project.services.logging().choices()
        self.fields["service"].required = False
        if len(self.fields["service"].choices) > 1 and not self.request.POST.get(
            "modal-service_title"
        ):
            self.hide_new_service = True
            self.fields["service"].widget.attrs["autofocus"] = True
        else:
            self.fields["service_title"].widget.attrs["autofocus"] = True

        if self.instance.pk:
            self.fields.pop("service_title")
            self.fields.pop("service_description")
            self.fields.pop("service_role")
            if (
                self.instance.rendered_by.enforce_same_week_logging
                and self.instance.rendered_on < logbook_lock()
            ):
                self.fields["hours"].disabled = True
                self.fields["rendered_by"].disabled = True
                self.fields["rendered_on"].disabled = True

        else:
            if self.request.user.features[FEATURES.GLASSFROG]:
                self.fields["service_role"].choices = Role.objects.choices()
            else:
                self.fields.pop("service_role")

    def clean(self):
        data = super().clean()
        errors = {}
        if not data.get("service") and not data.get("service_title"):
            errors["service"] = _(
                "This field is required unless you create a new service."
            )
        elif data.get("service") and data.get("service_title"):
            errors["service"] = _(
                "Deselect the existing service if you want to create a new service."
            )
        if self.project.closed_on:
            if self.project.is_logbook_locked:
                errors["__all__"] = _("This project has been closed too long ago.")
            else:
                self.add_warning(
                    _("This project has been closed recently."), code="project-closed"
                )
        if self.instance.invoice_service:
            self.add_warning(
                _("This entry is already part of an invoice."), code="part-of-invoice"
            )
        if data.get("service_title") and not is_title_specific(data["service_title"]):
            self.add_warning(
                _(
                    "This title seems awfully unspecific."
                    " Please use specific titles for services."
                ),
                code="unspecific-service",
            )
        if (
            data.get("service_title")
            and "service_role" in self.fields
            and not data.get("service_role")
        ):
            self.add_warning(_("No role selected."), code="no-role-selected")

        if all(
            f in self.fields and data.get(f) for f in ["rendered_by", "rendered_on"]
        ) and (not self.instance.pk or ("rendered_on" in self.changed_data)):
            if not data["rendered_by"].enforce_same_week_logging:
                # Fine
                pass
            elif data["rendered_on"] < logbook_lock():
                errors["rendered_on"] = _("Hours have to be logged in the same week.")
            elif data["rendered_on"] > in_days(7):
                errors["rendered_on"] = _("That's too far in the future.")

        if (
            all(data.get(f) for f in ["rendered_by", "rendered_on", "hours"])
            and not self.instance.pk
        ):
            msg = data["rendered_by"].take_a_break_warning(
                day=data["rendered_on"], add=data["hours"]
            )
            if msg:
                self.add_warning(msg, code="take-a-break")

        try:
            latest = LoggedHours.objects.filter(
                Q(rendered_by=self.request.user), ~Q(id=self.instance.id)
            ).latest("pk")
        except LoggedHours.DoesNotExist:
            pass
        else:
            fields = ["rendered_by", "rendered_on", "service", "hours", "description"]
            for field in fields:
                if data.get(field) != getattr(latest, field):
                    break
            else:
                self.add_warning(
                    _("This seems to be a duplicate. Is it?"), code="maybe-duplicate"
                )

        raise_if_errors(errors)
        return data

    def save(self):
        instance = super().save(commit=False)
        timestamp = None
        if not instance.pk:
            instance.created_by = self.request.user
            if pk := self.request.GET.get("timestamp"):
                timestamp = Timestamp.objects.filter(
                    user=self.request.user, id=pk
                ).first()
                timestamp.logged_hours = instance
            else:
                timestamp = DetectedTimestampForm(self.request.GET).build_if_valid(
                    user=self.request.user, logged_hours=instance
                )

        if not self.cleaned_data.get("service") and self.cleaned_data.get(
            "service_title"
        ):
            service = Service(
                project=self.project,
                title=self.cleaned_data["service_title"],
                description=self.cleaned_data["service_description"],
                role=self.cleaned_data.get("service_role"),
            )
            if self.project.flat_rate is not None:
                with override(settings.WORKBENCH.PDF_LANGUAGE):
                    service.effort_type = gettext("flat rate")
                    service.effort_rate = self.project.flat_rate
            service.save()
            instance.service = service
        instance.save()
        if timestamp:
            timestamp.save()

        return instance


@add_prefix("modal")
class LoggedCostForm(ModelForm):
    user_fields = default_to_current_user = ("rendered_by",)

    class Meta:
        model = LoggedCost
        fields = (
            "service",
            "rendered_by",
            "rendered_on",
            "expense_currency",
            "expense_cost",
            "third_party_costs",
            "are_expenses",
            "cost",
            "description",
        )
        widgets = {"description": Textarea({"rows": 2})}

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        if self.project:
            kwargs.setdefault("initial", {}).setdefault(
                "service", kwargs["request"].GET.get("service")
            )
        else:
            self.project = kwargs["instance"].service.project

        super().__init__(*args, **kwargs)
        if not self.instance.pk:
            self.instance.created_by = self.request.user

        self.fields["service"].choices = self.project.services.logging().choices()
        self.fields["cost"].help_text = mark_safe(
            "{} "
            '<a href="#" data-multiply-cost="1" class="">100%</a> '
            '<a href="#" data-multiply-cost="1.15" class="">115%</a> '
            "{}"
            "".format(
                _("Copy value from third party costs"), self.fields["cost"].help_text
            )
        )

        if self.instance.expense_report:
            self.fields["rendered_by"].disabled = True
            self.fields["rendered_on"].disabled = True
            self.fields["are_expenses"].disabled = True
            self.fields["third_party_costs"].disabled = True
            self.fields["expense_currency"].disabled = True
            self.fields["expense_cost"].disabled = True

        if self.request.user.features[FEATURES.FOREIGN_CURRENCIES]:
            rates = ExchangeRates.objects.newest()
            self.fields["expense_currency"] = forms.ChoiceField(
                choices=[("", "----------")]
                + [(currency, currency) for currency in rates.rates["rates"]],
                widget=forms.Select(attrs={"class": "custom-select"}),
                required=False,
                initial=self.instance.expense_currency,
                label=self.instance._meta.get_field("expense_currency").verbose_name,
            )
        else:
            self.fields.pop("expense_currency")
            self.fields.pop("expense_cost")

    def clean(self):
        data = super().clean()
        errors = {}
        if self.project.closed_on:
            if self.project.closed_on < in_days(-14):
                errors["__all__"] = _("This project has been closed too long ago.")
            else:
                self.add_warning(
                    _("This project has been closed recently."), code="project-closed"
                )
        if self.instance.invoice_service:
            self.add_warning(
                _("This entry is already part of an invoice."), code="part-of-invoice"
            )
        if data.get("are_expenses") and not data.get("third_party_costs"):
            errors["third_party_costs"] = (
                _("Providing third party costs is necessary for expenses."),
            )
        if data.get("cost") and data.get("third_party_costs") is not None:
            if data["cost"] < data["third_party_costs"]:
                self.add_warning(
                    _("Third party costs shouldn't be higher than costs."),
                    code="third-party-costs-higher",
                )
        if data.get("rendered_on") and data["rendered_on"] > in_days(7):
            errors["rendered_on"] = _("That's too far in the future.")
        raise_if_errors(errors)
        return data


class BreakSearchForm(Form):
    user = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["user"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        # data = self.cleaned_data
        # queryset = queryset.search(data.get("q"))
        queryset = self.apply_owned_by(queryset, attribute="user")
        return queryset.select_related("user")

    def response(self, request, queryset):
        if (
            request.GET.get("export") == "xlsx"
            and request.user.features[FEATURES.CONTROLLING]
        ):
            xlsx = WorkbenchXLSXDocument()
            xlsx.table_from_queryset(queryset)
            return xlsx.to_response("breaks.xlsx")


@add_prefix("modal")
class BreakForm(ModelForm):
    day = forms.DateField(label=capfirst(_("day")))
    starts_at = forms.TimeField(label=capfirst(_("starts at")))
    ends_at = forms.TimeField(label=capfirst(_("ends at")))

    class Meta:
        model = Break
        fields = ["description"]
        widgets = {"description": Textarea({"rows": 2})}

    def __init__(self, *args, **kwargs):
        request = kwargs["request"]
        initial = kwargs.setdefault("initial", {})
        for field in ["day", "starts_at", "ends_at", "description"]:
            if value := request.GET.get(field):
                initial[field] = value

        if "instance" not in kwargs:
            if "starts_at" not in initial:
                latest = localtime(request.user.latest_created_at)
                if dt.date.today() == latest.date():
                    initial.setdefault("starts_at", latest.time())
            initial.setdefault("day", dt.date.today())
            initial.setdefault("ends_at", localtime(timezone.now()).time())

        else:
            initial.setdefault("day", kwargs["instance"].starts_at.date())
            initial.setdefault(
                "starts_at", localtime(kwargs["instance"].starts_at).time()
            )
            initial.setdefault("ends_at", localtime(kwargs["instance"].ends_at).time())

        super().__init__(*args, **kwargs)

        self.order_fields(("day", "starts_at", "ends_at"))

    def clean(self):
        data = super().clean()
        errors = {}
        if data.get("day"):
            if data["day"] < logbook_lock() - dt.timedelta(days=7):
                errors["day"] = _("Breaks have to be logged promptly.")
            elif data["day"] > in_days(7):
                errors["day"] = _("That's too far in the future.")

        raise_if_errors(errors)

        if all(data.get(f) for f in ("day", "starts_at", "ends_at")):
            data["starts_at"] = timezone.make_aware(
                dt.datetime.combine(data["day"], data["starts_at"])
            )
            data["ends_at"] = timezone.make_aware(
                dt.datetime.combine(data["day"], data["ends_at"])
            )

        return data

    def save(self):
        new = not self.instance.pk
        instance = super().save(commit=False)
        instance.starts_at = self.cleaned_data["starts_at"]
        instance.ends_at = self.cleaned_data["ends_at"]
        if new:
            instance.user = self.request.user
        instance.save()

        if not new:
            return instance

        timestamp = None
        if pk := self.request.GET.get("timestamp"):
            Timestamp.objects.filter(user=self.request.user, id=pk).update(
                logged_break=instance
            )
        else:
            timestamp = DetectedTimestampForm(self.request.GET).build_if_valid(
                user=self.request.user, logged_break=instance
            )

        if timestamp:
            timestamp.save()
        return instance


class LoggedMoveForm(ModelForm):
    service = forms.ModelChoiceField(
        queryset=Service.objects.logging().filter(project__closed_on__isnull=True),
        widget=Autocomplete(model=Service),
        label="",
        initial="",
    )

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("initial", {}).setdefault("service", "")
        super().__init__(*args, **kwargs)

        if self.instance.archived_at:
            self.move_forbidden = _("Cannot move archived logbook entries.")
        elif self.instance.service.project.closed_on:
            self.move_forbidden = _("Cannot move logbook entries of closed projects.")
        else:
            self.move_forbidden = False
        if self.move_forbidden:
            messages.error(self.request, self.move_forbidden)

    def clean(self):
        data = super().clean()
        if self.move_forbidden:
            self.add_error("__all__", self.move_forbidden)
        return data


@add_prefix("modal")
class LoggedHoursMoveForm(LoggedMoveForm):
    class Meta:
        model = LoggedHours
        fields = ["service"]


@add_prefix("modal")
class LoggedCostMoveForm(LoggedMoveForm):
    class Meta:
        model = LoggedCost
        fields = ["service"]



# Generated by Django 3.1.5 on 2021-01-11 08:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("circles", "0002_auto_20190916_0843"),
    ]

    operations = [
        migrations.AddField(
            model_name="role",
            name="work_category",
            field=models.CharField(
                blank=True,
                choices=[
                    ("paid-work", "Paid work"),
                    ("knowledge-transfer", "Knowledge transfer"),
                    ("social-care", "Social care"),
                    ("outreach", "Outreach"),
                    ("other", "Other internal matters"),
                ],
                max_length=20,
                verbose_name="category",
            ),
        ),
    ]



# Generated by Django 2.2.2 on 2019-07-03 06:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("audit", "0002_auto_20190321_1626"),
        ("invoices", "0006_auto_20190309_2143"),
    ]

    operations = [
        migrations.RunSQL(
            "UPDATE audit_logged_actions SET row_data=delete(row_data, '_fts');"
            "UPDATE audit_logged_actions SET changed_fields=delete(changed_fields, '_fts');"
            "UPDATE audit_logged_actions SET row_data=delete(row_data, 'position');"
            "UPDATE audit_logged_actions SET changed_fields=delete(changed_fields, 'position');"
        ),
        migrations.RunSQL(
            "SELECT audit_audit_table('contacts_organization', ARRAY['fts_document']);"
            "SELECT audit_audit_table('contacts_person', ARRAY['fts_document', '_fts']);"
            "SELECT audit_audit_table('invoices_invoice', ARRAY['fts_document', '_fts']);"
            "SELECT audit_audit_table('invoices_service', ARRAY['position']);"
            "SELECT audit_audit_table('projects_project', ARRAY['fts_document', '_fts']);"
            "SELECT audit_audit_table('projects_service', ARRAY['position']);"
            "SELECT audit_audit_table('services_servicetype', ARRAY['position']);"
        ),
    ]



import datetime as dt
from collections import defaultdict
from decimal import Decimal
from functools import total_ordering

from django.contrib import messages
from django.db import models
from django.db.models import F, Prefetch, Q, Sum
from django.db.models.expressions import RawSQL
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.translation import gettext, gettext_lazy as _

from workbench.accounts.models import User
from workbench.contacts.models import Organization, Person
from workbench.services.models import ServiceBase
from workbench.tools.formats import Z1, Z2, local_date_format
from workbench.tools.models import Model, MoneyField, SearchQuerySet
from workbench.tools.urls import model_urls
from workbench.tools.validation import in_days, raise_if_errors


class CampaignQuerySet(SearchQuerySet):
    def open(self):
        return self.filter(
            Q(
                id__in=Project.objects.open()
                .filter(campaign__isnull=False)
                .values("campaign")
            )
        )

    def closed(self):
        return self.filter(
            ~Q(
                id__in=Project.objects.open()
                .filter(campaign__isnull=False)
                .values("campaign")
            )
        )


@model_urls
class Campaign(Model):
    customer = models.ForeignKey(
        Organization, on_delete=models.PROTECT, verbose_name=_("customer")
    )
    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"), blank=True)
    owned_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("responsible")
    )
    _fts = models.TextField(editable=False, blank=True)

    objects = CampaignQuerySet.as_manager()

    class Meta:
        ordering = ["-id"]
        verbose_name = _("campaign")
        verbose_name_plural = _("campaigns")

    def __str__(self):
        return "%s %s - %s" % (self.code, self.title, self.owned_by.get_short_name())

    def __html__(self):
        return format_html(
            "<small>{}</small> {} - {}",
            self.code,
            self.title,
            self.owned_by.get_short_name(),
        )

    def save(self, *args, **kwargs):
        self._fts = " ".join(str(part) for part in [self.code, self.customer.name])
        super().save(*args, **kwargs)

    save.alters_data = True

    @classmethod
    def allow_delete(cls, instance, request):
        return None

    @cached_property
    def statistics(self):
        from workbench.reporting.project_budget_statistics import (
            project_budget_statistics,
        )

        pbs = project_budget_statistics(self.projects.all())
        overall = pbs["overall"]
        overall["gross_margin_per_hour"] = (
            (overall["invoiced"] - overall["third_party_costs"]) / overall["hours"]
            if overall["hours"]
            else None
        )
        pbs["statistics"] = sorted(pbs["statistics"], key=lambda s: s["project"])
        return pbs


class ProjectQuerySet(SearchQuerySet):
    def open(self, *, on=None):
        return (
            self.filter(closed_on__isnull=True)
            if on is None
            else self.filter(
                Q(closed_on__isnull=True) | Q(closed_on__gt=on),
                created_at__lte=timezone.make_aware(
                    dt.datetime.combine(on, dt.time.max)
                ),
            )
        )

    def closed(self):
        return self.filter(closed_on__isnull=False)

    def orders(self):
        return self.filter(type=Project.ORDER)

    def without_invoices(self):
        from workbench.invoices.models import Invoice

        return self.exclude(
            id__in=Invoice.objects.invoiced()
            .filter(project__isnull=False)
            .values("project")
        )

    def with_accepted_offers(self):
        from workbench.offers.models import Offer

        return self.filter(id__in=Offer.objects.accepted().values("project"))

    def solely_declined_offers(self):
        from workbench.offers.models import Offer

        return self.filter(
            Q(id__in=Offer.objects.values("project"))
            & ~Q(
                id__in=Offer.objects.filter(
                    status__in=(Offer.IN_PREPARATION, Offer.OFFERED, Offer.ACCEPTED)
                ).values("project")
            )
        )

    def old_projects(self):
        from workbench.logbook.models import LoggedHours

        return (
            self.open()
            .filter(id__in=LoggedHours.objects.order_by().values("service__project"))
            .exclude(
                id__in=LoggedHours.objects.order_by()
                .filter(rendered_on__gte=in_days(-60))
                .values("service__project")
            )
        )

    def invalid_customer_contact_combination(self):
        return self.exclude(customer=F("contact__organization"))


@model_urls
@total_ordering
class Project(Model):
    ORDER = "order"
    MAINTENANCE = "maintenance"
    INTERNAL = "internal"

    TYPE_CHOICES = [
        (ORDER, _("Order")),
        (MAINTENANCE, _("Maintenance")),
        (INTERNAL, _("Internal")),
    ]

    customer = models.ForeignKey(
        Organization, on_delete=models.PROTECT, verbose_name=_("customer")
    )
    contact = models.ForeignKey(
        Person,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("contact"),
    )

    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(
        _("project description"),
        blank=True,
        help_text=_(
            "Do not use this for the offer description."
            " You can add the offer description later."
        ),
    )
    owned_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("responsible")
    )

    type = models.CharField(_("type"), choices=TYPE_CHOICES, max_length=20)
    flat_rate = MoneyField(
        _("flat rate"),
        blank=True,
        null=True,
        help_text=_("Set this if you want all services to have the same hourly rate."),
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    closed_on = models.DateField(_("closed on"), blank=True, null=True)

    _code = models.IntegerField(_("code"))
    _fts = models.TextField(editable=False, blank=True)

    cost_center = models.ForeignKey(
        "reporting.CostCenter",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("cost center"),
        related_name="projects",
    )
    campaign = models.ForeignKey(
        Campaign,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("campaign"),
        related_name="projects",
    )

    objects = ProjectQuerySet.as_manager()

    class Meta:
        ordering = ("-id",)
        verbose_name = _("project")
        verbose_name_plural = _("projects")

    def __str__(self):
        return "%s %s - %s" % (self.code, self.title, self.owned_by.get_short_name())

    def __html__(self):
        return format_html(
            "<small>{}</small> {} - {}",
            self.code,
            self.title,
            self.owned_by.get_short_name(),
        )

    def __lt__(self, other):
        return self.id < other.id if isinstance(other, Project) else 1

    @property
    def code(self):
        return "%s-%04d" % (self.created_at.year, self._code)

    def save(self, *args, **kwargs):
        new = not self.pk
        if new:
            self._code = RawSQL(
                "SELECT COALESCE(MAX(_code), 0) + 1 FROM projects_project"
                " WHERE EXTRACT(year FROM created_at) = %s",
                (timezone.now().year,),
            )
            super().save(*args, **kwargs)
            self.refresh_from_db()

        self._fts = " ".join(
            str(part)
            for part in [
                self.code,
                self.customer.name,
                self.contact.full_name if self.contact else "",
            ]
        )
        if new:
            super().save()
        else:
            super().save(*args, **kwargs)

    save.alters_data = True

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        errors = {}
        if self.closed_on and self.closed_on > dt.date.today():
            errors["closed_on"] = _(
                "Leave this empty if you do not want to close the project yet."
            )
        raise_if_errors(errors)

    @property
    def status_badge(self):
        css = {
            self.MAINTENANCE: "secondary",
            self.ORDER: "success",
            self.INTERNAL: "info",
        }[self.type]

        if self.closed_on:
            css = "light"
        return format_html(
            '<span class="badge badge-{}">{}</span>', css, self.pretty_status
        )

    @property
    def pretty_status(self):
        parts = [str(self.get_type_display())]
        if self.closed_on:
            parts.append(gettext("closed on %s") % local_date_format(self.closed_on))
        return ", ".join(parts)

    @cached_property
    def grouped_services(self):
        # Avoid circular imports
        from workbench.deals.models import Deal
        from workbench.logbook.models import LoggedCost, LoggedHours

        # Logged vs. service hours
        service_hours = defaultdict(lambda: Z1)
        logged_hours = defaultdict(lambda: Z1)
        # Logged vs. service cost
        service_cost = defaultdict(lambda: Z2)
        logged_cost = defaultdict(lambda: Z2)
        # Project logbook vs. project service cost (hours and cost)
        total_service_cost = Z2
        total_logged_cost = Z2
        total_service_hours_rate_undefined = Z1
        total_logged_hours_rate_undefined = Z1

        offers = self.offers.select_related("owned_by").prefetch_related(
            Prefetch("deals", Deal.objects.select_related("owned_by"))
        )
        offers_map = {offer.id: offer for offer in offers}
        services_by_offer = defaultdict(
            lambda: {"services": []}, ((offer, {"services": []}) for offer in offers)
        )

        logged_hours_per_service_and_user = defaultdict(dict)
        logged_hours_per_user = defaultdict(lambda: Z1)
        logged_hours_per_effort_rate = defaultdict(lambda: Z1)

        for row in (
            LoggedHours.objects.order_by()
            .filter(service__project=self)
            .values("service", "rendered_by")
            .annotate(Sum("hours"))
        ):
            logged_hours_per_user[row["rendered_by"]] += row["hours__sum"]
            logged_hours_per_service_and_user[row["service"]][row["rendered_by"]] = row[
                "hours__sum"
            ]

        logged_cost_per_service = {
            row["service"]: row["cost__sum"]
            for row in LoggedCost.objects.order_by()
            .filter(service__project=self)
            .values("service")
            .annotate(Sum("cost"))
        }

        not_archived_logged_hours_per_service = {
            row["service"]: row["hours__sum"]
            for row in LoggedHours.objects.order_by()
            .filter(
                service__project=self,
                archived_at__isnull=True,
                service__effort_rate__isnull=False,
            )
            .values("service")
            .annotate(Sum("hours"))
        }
        not_archived_logged_cost_per_service = {
            row["service"]: row["cost__sum"]
            for row in LoggedCost.objects.order_by()
            .filter(service__project=self, archived_at__isnull=True)
            .values("service")
            .annotate(Sum("cost"))
        }

        users = {
            user.id: user
            for user in User.objects.filter(id__in=logged_hours_per_user.keys())
        }

        for service in self.services.all():
            service.offer = offers_map.get(service.offer_id)  # Reuse
            logged = logged_hours_per_service_and_user.get(service.id, {})
            row = {
                "service": service,
                "logged_hours": sum(logged.values(), Z1),
                "logged_hours_per_user": sorted(
                    ((users[user], hours) for user, hours in logged.items()),
                    key=lambda row: row[1],
                    reverse=True,
                ),
                "logged_cost": logged_cost_per_service.get(service.id, Z2),
                "not_archived_logged_hours": not_archived_logged_hours_per_service.get(
                    service.id, Z1
                ),
                "not_archived_logged_cost": not_archived_logged_cost_per_service.get(
                    service.id, Z1
                ),
            }
            row["not_archived"] = (service.effort_rate or Z2) * row[
                "not_archived_logged_hours"
            ] + row["not_archived_logged_cost"]

            logged_hours[service.offer] += row["logged_hours"]
            logged_cost[service.offer] += row["logged_cost"]
            logged_hours[service.project] += row["logged_hours"]
            logged_cost[service.project] += row["logged_cost"]
            total_logged_cost += row["logged_cost"]
            logged_hours_per_effort_rate[service.effort_rate] += row["logged_hours"]

            if not service.is_declined:
                service_hours[service.offer] += service.service_hours
                service_cost[service.offer] += service.cost or Z2
                service_hours[service.project] += service.service_hours
                service_cost[service.project] += service.cost or Z2
                total_service_cost += service.service_cost

            if service.effort_rate is not None:
                total_logged_cost += service.effort_rate * row["logged_hours"]
            else:
                total_logged_hours_rate_undefined += row["logged_hours"]
                if not service.is_declined:
                    total_service_hours_rate_undefined += service.service_hours

            services_by_offer[offers_map.get(service.offer_id)]["services"].append(row)

        for offer, offer_data in services_by_offer.items():
            offer_data["service_hours"] = service_hours[offer]
            offer_data["logged_hours"] = logged_hours[offer]
            offer_data["service_cost"] = service_cost[offer]
            offer_data["logged_cost"] = logged_cost[offer]

        return {
            "offers": sorted(
                item
                for item in services_by_offer.items()
                if item[1]["services"] or item[0] is not None
            ),
            "logged_hours": logged_hours[self],
            "logged_hours_per_user": sorted(
                ((users[user], hours) for user, hours in logged_hours_per_user.items()),
                key=lambda row: row[1],
                reverse=True,
            ),
            "logged_hours_per_effort_rate": sorted(
                (
                    (rate, hours)
                    for rate, hours in logged_hours_per_effort_rate.items()
                    if hours
                ),
                key=lambda row: row[0] or Decimal("9999999"),
                reverse=True,
            ),
            "logged_cost": logged_cost[self],
            "service_hours": service_hours[self],
            "service_cost": service_cost[self],
            "total_service_cost": total_service_cost,
            "total_logged_cost": total_logged_cost,
            "total_service_hours_rate_undefined": total_service_hours_rate_undefined,
            "total_logged_hours_rate_undefined": total_logged_hours_rate_undefined,
            "total_discount": sum(
                (offer.discount for offer in offers if not offer.is_declined), Z2
            ),
        }

    @cached_property
    def project_invoices(self):
        return self.invoices.select_related("contact__organization").reverse()

    @cached_property
    def project_invoices_total_excl_tax(self):
        return sum(
            (
                invoice.total_excl_tax
                for invoice in self.project_invoices
                if invoice.is_invoiced
            ),
            Z2,
        )

    @cached_property
    def not_archived_total(self):
        # Avoid circular imports
        from workbench.logbook.models import LoggedCost, LoggedHours

        total = Z2
        hours_rate_undefined = Z1

        for row in (
            LoggedHours.objects.order_by()
            .filter(service__project=self, archived_at__isnull=True)
            .values("service__effort_rate")
            .annotate(Sum("hours"))
        ):
            if row["service__effort_rate"] is None:
                hours_rate_undefined += row["hours__sum"]
            else:
                total += row["hours__sum"] * row["service__effort_rate"]

        total += (
            LoggedCost.objects.order_by()
            .filter(service__project=self, archived_at__isnull=True)
            .aggregate(Sum("cost"))["cost__sum"]
            or Z2
        )
        return {"total": total, "hours_rate_undefined": hours_rate_undefined}

    def solely_declined_offers_warning(self, *, request):
        from workbench.offers.models import Offer

        if self.closed_on:
            return

        status = set(self.offers.order_by().values_list("status", flat=True))
        if status == {Offer.DECLINED}:
            messages.warning(
                request,
                _(
                    "All offers of project %(project)s are declined."
                    " You might want to close the project now?"
                )
                % {"project": self},
            )

    @property
    def is_logbook_locked(self):
        return self.closed_on and self.closed_on < in_days(-14)


class ServiceQuerySet(SearchQuerySet):
    def choices(self):
        offers = defaultdict(list)
        for service in self.select_related("offer__project", "offer__owned_by"):
            offers[service.offer].append((service.id, str(service)))
        return [("", "----------")] + [
            (offer or _("Not offered yet"), services)
            for offer, services in sorted(
                offers.items(),
                key=lambda item: (
                    item[0] and item[0].offered_on or dt.date.max,
                    item[0] and item[0].pk or 1e100,
                ),
            )
        ]

    def budgeted(self):
        from workbench.offers.models import Offer

        return self.filter(Q(offer__isnull=True) | ~Q(offer__status=Offer.DECLINED))

    def logging(self):
        return self.budgeted().filter(allow_logging=True)

    def editable(self):
        from workbench.offers.models import Offer

        return self.filter(
            Q(offer__isnull=True) | Q(offer__status=Offer.IN_PREPARATION)
        )


@model_urls
class Service(ServiceBase):
    RELATED_MODEL_FIELD = "offer"

    project = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name="services",
        verbose_name=_("project"),
    )
    offer = models.ForeignKey(
        "offers.Offer",
        on_delete=models.SET_NULL,
        related_name="services",
        verbose_name=_("offer"),
        blank=True,
        null=True,
    )
    allow_logging = models.BooleanField(
        _("allow logging"),
        default=True,
        help_text=_(
            "Deactivate this for service entries which are only used for budgeting."
        ),
    )
    is_optional = models.BooleanField(
        _("is optional"),
        default=False,
        help_text=_("Optional services to not count towards the offer total."),
    )
    role = models.ForeignKey(
        "circles.Role",
        on_delete=models.PROTECT,
        related_name="services",
        verbose_name=_("role"),
        blank=True,
        null=True,
    )

    objects = ServiceQuerySet.as_manager()

    def get_absolute_url(self):
        return "%s#service%s" % (self.project.get_absolute_url(), self.pk)

    @classmethod
    def allow_update(cls, instance, request):
        return True

    @classmethod
    def allow_delete(cls, instance, request):
        if instance.offer and instance.offer.status > instance.offer.IN_PREPARATION:
            messages.error(
                request,
                _(
                    "Cannot delete a service bound to an offer"
                    " which is not in preparation anymore."
                ),
            )
            return False
        return super().allow_delete(instance, request)

    @classmethod
    def get_redirect_url(cls, instance, request):
        if not request.is_ajax():
            return instance.get_absolute_url() if instance else "projects_project_list"

    @property
    def is_declined(self):
        return self.offer.is_declined if self.offer else False



# Generated by Django 2.2.2 on 2019-07-09 19:38

from decimal import Decimal

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="ExpenseReport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "total",
                    models.DecimalField(
                        blank=True,
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="Total incl. tax for third-party services.",
                        max_digits=10,
                        null=True,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="total",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created by",
                    ),
                ),
                (
                    "owned_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="expensereports",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="responsible",
                    ),
                ),
            ],
            options={
                "verbose_name": "expense report",
                "verbose_name_plural": "expense reports",
                "ordering": ("-created_at",),
            },
        )
    ]



from collections import defaultdict

from django.db.models import Sum

from workbench.accounts.models import User
from workbench.contacts.models import Organization
from workbench.logbook.models import LoggedHours
from workbench.projects.models import Project, Service
from workbench.tools.formats import Z1


def overdrawn_projects():
    projects = (
        Project.objects.open()
        .exclude(type=Project.INTERNAL)
        .select_related("customer", "owned_by")
    )

    logged_hours = {
        row["service__project"]: row["hours__sum"]
        for row in LoggedHours.objects.order_by()
        .filter(service__project__in=projects)
        .values("service__project")
        .annotate(Sum("hours"))
    }

    service_hours = {
        row["project"]: row["service_hours__sum"]
        for row in Service.objects.order_by()
        .filter(project__in=projects)
        .values("project")
        .annotate(Sum("service_hours"))
    }

    projects = [
        {
            "project": project,
            "logged_hours": logged_hours.get(project.id, Z1),
            "service_hours": service_hours.get(project.id, Z1),
            "delta": logged_hours.get(project.id, Z1)
            - service_hours.get(project.id, Z1),
        }
        for project in projects
    ]

    TYPE_ORDERING = {Project.ORDER: 1, Project.MAINTENANCE: 2}

    return sorted(
        (
            project
            for project in projects
            if project["logged_hours"] > project["service_hours"]
        ),
        key=lambda row: (TYPE_ORDERING.get(row["project"].type, 9), -row["delta"]),
    )


def hours_per_customer(date_range, *, users=None):
    hours = defaultdict(lambda: defaultdict(lambda: Z1))
    user_hours = defaultdict(lambda: Z1)
    seen_organizations = set()
    seen_users = set()

    queryset = LoggedHours.objects.order_by().filter(rendered_on__range=date_range)
    if users:
        queryset = queryset.filter(rendered_by__in=users)

    for row in queryset.values("rendered_by", "service__project__customer").annotate(
        Sum("hours")
    ):
        hours[row["service__project__customer"]][row["rendered_by"]] = row["hours__sum"]
        user_hours[row["rendered_by"]] += row["hours__sum"]
        seen_organizations.add(row["service__project__customer"])
        seen_users.add(row["rendered_by"])

    organizations = []
    user_list = User.objects.filter(id__in=seen_users)

    for org in Organization.objects.filter(id__in=seen_organizations):
        organizations.append(
            {
                "organization": org,
                "user_hours": [(user, hours[org.id][user.id]) for user in user_list],
                "total_hours": sum(hours[org.id].values(), Z1),
            }
        )

    return {
        "organizations": sorted(
            organizations, key=lambda row: row["total_hours"], reverse=True
        ),
        "users": user_list,
        "user_hours": [(user, user_hours[user.id]) for user in user_list],
        "total_hours": sum(user_hours.values(), Z1),
    }



import io
import json
import re
from decimal import Decimal
from html.parser import HTMLParser

from django.core.management import BaseCommand
from django.utils.dateparse import parse_date, parse_datetime
from django.utils.translation import activate

from workbench.accounts.middleware import set_user_name
from workbench.offers.models import Offer
from workbench.projects.models import Project, Service


class _DeHTMLParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.__text = []

    def handle_data(self, data):
        text = data.strip()
        if len(text) > 0:
            text = re.sub("[ \t\r\n]+", " ", text)
            self.__text.append(text + " ")

    def handle_entityref(self, name):
        self.__text.append(self.unescape("&%s;" % name))

    def handle_charref(self, name):
        self.__text.append(self.unescape("&#%s;" % name))

    def handle_starttag(self, tag, attrs):
        if tag == "p":
            self.__text.append("\n\n")
        elif tag == "br":
            self.__text.append("\n")
        elif tag == "li":
            self.__text.append("\n- ")

    def handle_startendtag(self, tag, attrs):
        if tag == "br":
            self.__text.append("\n\n")

    def text(self):
        return "".join(self.__text).strip()


def dehtml(text):
    try:
        parser = _DeHTMLParser()
        parser.feed(text)
        parser.close()
        return parser.text()
    except Exception:
        return text


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("dump", nargs=1)

    def handle(self, **options):
        activate("de")
        set_user_name("Offer importer")

        with io.open(options["dump"][0]) as f:
            data = json.load(f)

        offers = {}

        contacts = {
            row["pk"]: {
                "customer_id": row["fields"]["employer"],
                "contact_id": row["fields"]["person"],
            }
            for row in data
            if row["model"] == "addressbook.address"
        }
        activities = {
            row["pk"]: row["fields"]
            for row in data
            if row["model"] == "articles.activity"
        }
        articles = {
            row["pk"]: row["fields"]
            for row in data
            if row["model"] == "articles.article"
        }

        def article_name(stuff, pk):
            if stuff[pk]["level"] <= 1:
                return stuff[pk]["name"]
            else:
                return "%s - %s" % (
                    stuff[stuff[pk]["parent"]]["name"],
                    stuff[pk]["name"],
                )

        for row in data:
            if row["model"] == "offers.joboffer":
                if row["pk"] not in offers:
                    offers[row["pk"]] = {
                        "project": Project(
                            title=row["fields"]["name"],
                            description="",
                            owned_by_id=row["fields"]["manager"],
                            type=Project.ORDER,
                            created_at=parse_datetime(row["fields"]["created"]),
                            **contacts[row["fields"]["contact"]]
                        ),
                        "offer": Offer(
                            title=row["fields"]["name"],
                            description=dehtml(row["fields"]["description"]),
                            owned_by_id=row["fields"]["manager"],
                            created_at=parse_datetime(row["fields"]["created"]),
                            offered_on=parse_date(row["fields"]["offer_date"]),
                            status=Offer.OFFERED,
                            postal_address="%s\r\n%s %s\r\n%s\r\n%s %s"
                            % (
                                row["fields"]["offer_company"],
                                row["fields"]["offer_first_name"],
                                row["fields"]["offer_last_name"],
                                row["fields"]["offer_address"],
                                row["fields"]["offer_zip_code"],
                                row["fields"]["offer_city"],
                            ),
                            subtotal=Decimal(row["fields"]["subtotal"]),
                            discount=Decimal(row["fields"]["discount"]),
                            liable_to_vat=True,
                            tax_rate=Decimal(row["fields"]["tax"]),
                            total=Decimal(row["fields"]["total"]),
                        ),
                        "services": [],
                    }

            elif row["model"] == "offers.offeredactivity":
                offers[row["fields"]["joboffer"]]["services"].append(
                    Service(
                        created_at=parse_datetime(row["fields"]["created"]),
                        title=article_name(activities, row["fields"]["activity"]),
                        description=row["fields"]["notes"],
                        position=row["fields"]["ordering"],
                        effort_type=article_name(activities, row["fields"]["activity"]),
                        effort_hours=Decimal(row["fields"]["hours"]),
                        effort_rate=Decimal(row["fields"]["invoicing_hourly_rate"]),
                        is_optional=row["fields"]["optional"],
                    )
                )

            elif row["model"] == "offers.offeredarticle":
                offers[row["fields"]["joboffer"]]["services"].append(
                    Service(
                        created_at=parse_datetime(row["fields"]["created"]),
                        title=article_name(articles, row["fields"]["article"]),
                        description=row["fields"]["notes"],
                        position=1000 + row["fields"]["ordering"],
                        cost=row["fields"]["count"]
                        * Decimal(row["fields"]["cost"])
                        * (1 + Decimal(row["fields"]["commission"]) / 100),
                    )
                )

        for offer in offers.values():
            offer["project"].save()
            offer["offer"].project = offer["project"]
            offer["offer"].save()
            for service in offer["services"]:
                service.project = offer["project"]
                service.offer = offer["offer"]
                service.save(skip_related_model=True)

            offer["offer"].save()



import datetime as dt
from decimal import Decimal

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.test.utils import override_settings
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.invoices.models import Invoice
from workbench.tools.formats import local_date_format
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days


def invoice_to_dict(invoice, **kwargs):
    return {
        "customer": invoice.customer_id or "",
        "contact": invoice.contact_id or "",
        "title": invoice.title,
        "description": invoice.description,
        "owned_by": invoice.owned_by_id,
        "subtotal": invoice.subtotal,
        "discount": invoice.discount,
        "third_party_costs": invoice.third_party_costs,
        "liable_to_vat": invoice.liable_to_vat,
        "postal_address": invoice.postal_address,
        "status": invoice.status,
        "type": invoice.type,
        "closed_on": invoice.closed_on and invoice.closed_on.isoformat() or "",
        "invoiced_on": invoice.invoiced_on and invoice.invoiced_on.isoformat() or "",
        "due_on": invoice.due_on and invoice.due_on.isoformat() or "",
        "show_service_details": invoice.show_service_details,
        **kwargs,
    }


class InvoicesTest(TestCase):
    fixtures = ["exchangerates.json"]

    def setUp(self):
        deactivate_all()

    def test_down_payment_invoice(self):
        """Down payment invoices do not have service details and validate the
        postal address (as any other invoice too)"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        url = project.urls["createinvoice"] + "?type=down-payment"
        response = self.client.get(url)
        self.assertContains(response, "Down payment")
        self.assertNotContains(response, "id_show_service_details")

        response = self.client.post(
            url,
            {
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "discount": 0,
                "liable_to_vat": 1,
                "postal_address": "Street\nCity",
                "subtotal": 2500,
                "third_party_costs": 0,
            },
        )
        self.assertContains(response, 'value="short-postal-address"')

        response = self.client.post(
            url,
            {
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "discount": 0,
                "liable_to_vat": 1,
                "postal_address": "Anything\nStreet\nCity",
                "subtotal": 2500,
                "third_party_costs": 0,
            },
        )

        invoice = Invoice.objects.get()
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertAlmostEqual(invoice.subtotal, Decimal("2500"))

    def test_create_service_invoice_from_offer(self):
        """Service invoice creation from offers (services) works and results in
        the expected invoice total"""
        service = factories.ServiceFactory.create(cost=100, allow_logging=True)
        url = service.project.urls["createinvoice"] + "?type=services&source=offer"
        self.client.force_login(service.project.owned_by)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "id_show_service_details")

        response = self.client.post(
            url,
            {
                "contact": service.project.contact_id,
                "title": service.project.title,
                "owned_by": service.project.owned_by_id,
                "discount": "0",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "selected_services": [service.pk],
                "disable_logging": 1,
            },
        )
        # print(response, response.content.decode("utf-8"))

        invoice = Invoice.objects.get()
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertEqual(invoice.subtotal, 100)
        self.assertEqual(invoice.service_period, None)

        service.refresh_from_db()
        self.assertFalse(service.allow_logging)

        service = invoice.services.get()
        response = self.client.post(
            service.urls["update"],
            {
                "title": service.title,
                "description": service.description,
                "effort_type": service.effort_type,
                "effort_rate": service.effort_rate or "",
                "effort_hours": service.effort_hours or "",
                "cost": service.cost or "",
                "third_party_costs": service.third_party_costs or "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        self.assertRedirects(
            self.client.post(invoice.urls["delete"]), invoice.urls["list"]
        )
        self.assertEqual(Invoice.objects.count(), 0)

    def test_create_service_invoice_from_logbook(self):
        """Service invoice creation from the logbook works and results in the
        expected invoice total"""
        project = factories.ProjectFactory.create()
        service1 = factories.ServiceFactory.create(
            project=project, title="cost-only", cost=100
        )
        service2 = factories.ServiceFactory.create(project=project, title="no-rate")
        service3 = factories.ServiceFactory.create(
            project=project,
            title="with-rate",
            effort_type="Consulting",
            effort_rate=200,
        )
        service4 = factories.ServiceFactory.create(project=project, title="nothing")

        cost = factories.LoggedCostFactory.create(
            service=service1,
            cost=10,
            description="Test",
            rendered_on=dt.date(2020, 3, 18),
        )
        hours = factories.LoggedHoursFactory.create(
            service=service1,
            hours=1,
            description="Test",
            rendered_on=dt.date(2020, 3, 20),
        )
        factories.LoggedHoursFactory.create(
            service=service2, hours=2, rendered_on=dt.date(2020, 3, 20)
        )
        factories.LoggedHoursFactory.create(
            service=service3, hours=3, rendered_on=dt.date(2020, 3, 22)
        )

        url = project.urls["createinvoice"] + "?type=services&source=logbook"
        self.client.force_login(project.owned_by)
        response = self.client.get(url)
        # print(response, response.content.decode("utf-8"))

        self.assertContains(response, "<strong>cost-only</strong><br>10.00")
        self.assertContains(response, "1.0h logged but no hourly rate defined.")
        self.assertContains(response, "<strong>no-rate</strong><br>0.00")
        self.assertContains(response, "2.0h logged but no hourly rate defined.")
        self.assertContains(response, "<strong>with-rate</strong><br>600.00")
        self.assertContains(response, "id_show_service_details")

        cost.service = service1
        cost.save()

        response = self.client.post(
            url,
            {
                "contact": project.contact_id,
                "title": project.title,
                "owned_by": project.owned_by_id,
                "discount": "0",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "selected_services": [
                    service1.pk,
                    service2.pk,
                    service3.pk,
                    service4.pk,
                ],
                "disable_logging": 0,
            },
        )
        invoice = Invoice.objects.get()
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertEqual(invoice.subtotal, 610)
        self.assertEqual(invoice.service_period, "18.03.2020 - 22.03.2020")

        cost.refresh_from_db()
        self.assertEqual(cost.invoice_service.invoice, invoice)
        hours.refresh_from_db()
        self.assertEqual(hours.invoice_service.invoice, invoice)

        self.assertEqual(service1.invoice_services.get().invoice, invoice)
        self.assertEqual(service2.invoice_services.get().invoice, invoice)
        self.assertEqual(service3.invoice_services.get().invoice, invoice)
        self.assertEqual(service4.invoice_services.count(), 0)

        response = self.client.post(
            cost.urls["update"],
            {
                "service": cost.service_id,
                "rendered_on": cost.rendered_on.isoformat(),
                "third_party_costs": cost.third_party_costs or "",
                "cost": 2 * cost.cost,
                "description": cost.description,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "This entry is already part of an invoice.")

        response = self.client.post(
            hours.urls["update"],
            {
                "service": hours.service_id,
                "rendered_on": hours.rendered_on.isoformat(),
                "rendered_by": hours.rendered_by_id,
                "hours": hours.hours,
                "description": hours.description,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "This entry is already part of an invoice.")

        response = self.client.post(
            cost.urls["update"],
            {
                "modal-service": cost.service_id,
                "modal-rendered_by": cost.rendered_by_id,
                "modal-rendered_on": cost.rendered_on.isoformat(),
                "modal-third_party_costs": cost.third_party_costs or "",
                "modal-cost": 2 * cost.cost,
                "modal-description": cost.description,
                WarningsForm.ignore_warnings_id: "part-of-invoice",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        self.assertContains(
            self.client.get("/"),
            "Logged cost &#x27;Test&#x27; has been updated successfully.",
        )

        cost.refresh_from_db()
        self.assertAlmostEqual(cost.cost, Decimal("20"))
        invoice.refresh_from_db()
        self.assertAlmostEqual(invoice.subtotal, 610)  # unchanged

        response = self.client.get(invoice.urls["pdf"])
        self.assertEqual(response.status_code, 200)  # No crash

        response = self.client.get(invoice.urls["xlsx"])
        self.assertEqual(response.status_code, 200)  # No crash

        response = self.client.post(
            invoice.urls["delete"],
            {WarningsForm.ignore_warnings_id: "release-logged-services"},
        )
        self.assertRedirects(response, invoice.urls["list"])
        self.assertEqual(Invoice.objects.count(), 0)
        self.assertEqual(
            messages(response),
            ["Invoice '{}' has been deleted successfully.".format(invoice)],
        )

    def test_delete_service_invoice_with_logs(self):
        """Deleting service invoices with related logbook entries unarchives
        those entries"""
        service = factories.ServiceFactory.create()
        cost = factories.LoggedCostFactory.create(
            cost=150, service=service, description="this"
        )

        url = service.project.urls["createinvoice"] + "?type=services&source=logbook"
        self.client.force_login(service.project.owned_by)
        response = self.client.post(
            url,
            {
                "contact": service.project.contact_id,
                "title": service.project.title,
                "owned_by": service.project.owned_by_id,
                "discount": "0",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "selected_services": [service.pk],
                "disable_logging": 0,
            },
        )

        invoice = Invoice.objects.get()
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertAlmostEqual(invoice.subtotal, Decimal(150))

        cost.refresh_from_db()
        self.assertEqual(cost.invoice_service.invoice, invoice)
        self.assertEqual(cost.invoice_service.project_service, service)

        response = self.client.get(invoice.urls["delete"])
        self.assertContains(response, WarningsForm.ignore_warnings_id)

        response = self.client.post(invoice.urls["delete"])
        self.assertContains(response, "Logged services are linked with this invoice.")
        self.assertEqual(Invoice.objects.count(), 1)
        cost.refresh_from_db()
        self.assertTrue(cost.invoice_service)

        response = self.client.post(
            invoice.urls["delete"],
            {WarningsForm.ignore_warnings_id: "release-logged-services"},
        )
        self.assertRedirects(response, invoice.urls["list"])
        self.assertEqual(
            messages(response),
            ["Invoice '{}' has been deleted successfully.".format(invoice)],
        )

        cost.refresh_from_db()
        self.assertEqual(cost.invoice_service, None)

        # Creating the invoice again succeeds.
        response = self.client.post(
            url,
            {
                "contact": service.project.contact_id,
                "title": service.project.title,
                "owned_by": service.project.owned_by_id,
                "discount": "0",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "selected_services": [service.pk],
                "disable_logging": 0,
            },
        )
        invoice = Invoice.objects.get()
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertAlmostEqual(invoice.subtotal, Decimal(150))

    def test_pre_form(self):
        """Creating invoices directly shows a pre-form allowing selection of
        customer/contact combinations"""
        self.client.force_login(factories.UserFactory.create())

        # pre_form does not have these fields
        response = self.client.get(Invoice.urls["create"])
        self.assertContains(response, 'method="GET"')
        self.assertNotContains(response, 'id="id_title"')
        self.assertNotContains(response, 'id="id_description"')

        # Nonexistant entries
        response = self.client.get(Invoice.urls["create"] + "?contact=0")
        self.assertContains(response, 'method="GET"')
        self.assertNotContains(response, 'id="id_title"')
        self.assertNotContains(response, 'id="id_description"')

        response = self.client.get(Invoice.urls["create"] + "?customer=0")
        self.assertContains(response, 'method="GET"')
        self.assertNotContains(response, 'id="id_title"')
        self.assertNotContains(response, 'id="id_description"')

    def test_create_update_person_invoice(self):
        """Test creating and updating invoices not linked to projects"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )
        self.client.force_login(person.primary_contact)

        url = Invoice.urls["create"] + "?contact={}".format(person.pk)
        response = self.client.get(url)
        self.assertContains(response, 'method="POST"')
        self.assertNotContains(response, 'data-field-value="')
        postal_address = factories.PostalAddressFactory.create(person=person)
        response = self.client.get(url)
        self.assertContains(response, 'data-field-value="', 1)

        person.organization.default_billing_address = "Default"
        person.organization.save()
        response = self.client.get(url)
        self.assertContains(response, 'data-field-value="', 2)

        response = self.client.post(
            url,
            {
                "customer": person.organization_id,
                "contact": person.id,
                "title": "Stuff",
                "owned_by": person.primary_contact_id,
                "subtotal": "110",
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": postal_address.postal_address,
                "third_party_costs": 0,
            },
        )
        invoice = Invoice.objects.get()
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertAlmostEqual(invoice.total_excl_tax, Decimal("100"))
        self.assertAlmostEqual(invoice.total, Decimal("107.7"))

    def test_customer_create_invoice(self):
        """Creating invoices for organizations shows the postal address selector too"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )
        self.client.force_login(person.primary_contact)
        response = self.client.get(
            "/invoices/create/?customer={}".format(person.organization.id)
        )
        self.assertContains(
            response, 'value="The Organization Ltd" placeholder="Organization"'
        )
        self.assertContains(response, 'id="id_postal_address"')
        self.assertNotContains(response, 'data-field-value="')
        self.assertNotContains(response, "id_show_service_details")

        person.organization.default_billing_address = "Default"
        person.organization.save()

        response = self.client.get(
            "/invoices/create/?customer={}".format(person.organization.id)
        )
        self.assertContains(response, 'id="id_postal_address"')
        self.assertContains(response, 'data-field-value="')

    def test_update_invoice(self):
        """Updating invoices produces a variety of errors and warnings"""
        invoice = factories.InvoiceFactory.create(
            contact=None, postal_address="Test\nStreet\nCity"
        )
        self.client.force_login(invoice.owned_by)
        response = self.client.post(invoice.urls["update"], invoice_to_dict(invoice))
        self.assertContains(response, "No contact selected.")

        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, **{WarningsForm.ignore_warnings_id: "no-contact"}),
        )
        self.assertRedirects(response, invoice.urls["detail"])

        response = self.client.get(invoice.urls["delete"])
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            invoice.urls["update"], invoice_to_dict(invoice, status=Invoice.SENT)
        )
        self.assertContains(
            response, "Invoice and/or due date missing for selected state."
        )

        person = factories.PersonFactory.create(organization=invoice.customer)
        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(
                invoice,
                contact=person.id,
                status=Invoice.SENT,
                invoiced_on=dt.date.today().isoformat(),
                due_on=dt.date.today().isoformat(),
            ),
        )
        self.assertRedirects(response, invoice.urls["detail"])

        invoice.refresh_from_db()
        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, closed_on=dt.date.today().isoformat()),
        )
        self.assertContains(response, "Invalid status when closed on is already set.")

        response = self.client.get(invoice.urls["delete"])
        self.assertRedirects(response, invoice.urls["detail"])
        self.assertEqual(
            messages(response), ["Invoices in preparation may be deleted, others not."]
        )

        invoice.refresh_from_db()
        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, postal_address=invoice.postal_address + " hello"),
        )
        # print(response, response.content.decode("utf-8"))
        self.assertContains(
            response,
            "You are attempting to change &#x27;Postal address&#x27;."
            " I am trying to prevent unintentional changes. Are you sure?",
        )

        response = self.client.post(
            invoice.urls["update"], invoice_to_dict(invoice, status=Invoice.PAID)
        )
        self.assertRedirects(response, invoice.urls["detail"])

        invoice.refresh_from_db()
        self.assertEqual(invoice.closed_on, dt.date.today())

    def test_list(self):
        """Filter form smoke test"""
        factories.InvoiceFactory.create()
        user = factories.UserFactory.create()
        self.client.force_login(user)

        code = check_code(self, "/invoices/")
        code("")
        code("q=test")
        code("s=open")
        code("s=40")  # PAID
        code("org={}".format(factories.OrganizationFactory.create().pk))
        code("owned_by={}".format(user.id))
        code("owned_by=-1")  # mine
        code("owned_by=0")  # only inactive
        code("export=xlsx")

    @override_settings(BATCH_MAX_ITEMS=5)
    def test_too_many_invoices(self):
        """Creating a PDF with too many invoices fails"""
        invoice = factories.InvoiceFactory.create()

        for i in range(5):
            factories.InvoiceFactory.create(
                customer=invoice.customer,
                contact=invoice.contact,
                owned_by=invoice.owned_by,
            )

        self.client.force_login(invoice.owned_by)
        response = self.client.get("/invoices/?export=pdf")
        self.assertRedirects(
            response, "/invoices/?error=1", fetch_redirect_response=False
        )
        self.assertEqual(
            messages(response), ["6 invoices in selection, that's too many."]
        )

    def test_list_pdfs(self):
        """Various checks when exporting PDFs of lists"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get("/invoices/?export=pdf")
        self.assertEqual(response.status_code, 302)
        self.assertEqual(messages(response), ["No invoices found."])

        factories.InvoiceFactory.create(
            invoiced_on=in_days(-60),
            due_on=in_days(-45),
            status=Invoice.SENT,
        )
        response = self.client.get("/invoices/?export=pdf")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

    def test_model_validation(self):
        """Invoice model validation"""
        invoice = Invoice(
            title="Test",
            customer=factories.OrganizationFactory.create(),
            owned_by=factories.UserFactory.create(),
            type=Invoice.FIXED,
            _code=0,
            status=Invoice.SENT,
            postal_address="Test\nStreet\nCity",
        )
        msg = ["Invoice and/or due date missing for selected state."]

        with self.assertRaises(ValidationError) as cm:
            invoice.clean_fields(exclude=["status"])
        self.assertEqual(list(cm.exception), msg)

        with self.assertRaises(ValidationError) as cm:
            invoice.clean_fields()
        self.assertEqual(list(cm.exception), [("status", msg)])

        with self.assertRaises(ValidationError) as cm:
            Invoice(
                title="Test",
                customer=factories.OrganizationFactory.create(),
                owned_by=factories.UserFactory.create(),
                type=Invoice.FIXED,
                _code=0,
                status=Invoice.SENT,
                postal_address="Test\nStreet\nCity",
                invoiced_on=dt.date.today(),
                due_on=in_days(-1),
            ).full_clean()
        self.assertEqual(
            list(cm.exception), [("due_on", ["Due date has to be after invoice date."])]
        )

        with self.assertRaises(ValidationError) as cm:
            Invoice(
                title="Test",
                customer=factories.OrganizationFactory.create(),
                owned_by=factories.UserFactory.create(),
                type=Invoice.DOWN_PAYMENT,
                _code=0,
                status=Invoice.IN_PREPARATION,
                postal_address="Test\nStreet\nCity",
            ).full_clean()
        self.assertEqual(
            list(cm.exception),
            [("__all__", ["Invoices of type Down payment require a project."])],
        )

        with self.assertRaises(ValidationError) as cm:
            Invoice(
                title="Test",
                customer=factories.OrganizationFactory.create(),
                owned_by=factories.UserFactory.create(),
                type=Invoice.FIXED,
                _code=0,
                status=Invoice.SENT,
                postal_address="Test\nStreet\nCity",
                invoiced_on=dt.date.today(),
                due_on=in_days(0),
                service_period_from=in_days(0),
                service_period_until=None,
            ).full_clean()
        self.assertEqual(
            list(cm.exception),
            [
                ("service_period_from", ["Either fill in both fields or none."]),
                ("service_period_until", ["Either fill in both fields or none."]),
            ],
        )

        with self.assertRaises(ValidationError) as cm:
            Invoice(
                title="Test",
                customer=factories.OrganizationFactory.create(),
                owned_by=factories.UserFactory.create(),
                type=Invoice.FIXED,
                _code=0,
                status=Invoice.SENT,
                postal_address="Test\nStreet\nCity",
                invoiced_on=dt.date.today(),
                due_on=in_days(0),
                service_period_from=in_days(0),
                service_period_until=in_days(-1),
            ).full_clean()
        self.assertEqual(
            list(cm.exception),
            [("service_period_until", ["Until date has to be after from date."])],
        )

    def test_unlock_sent_invoice(self):
        """Unlocking sent invoices is possible after ignoring a warning"""
        invoice = factories.InvoiceFactory.create(
            title="Test",
            subtotal=20,
            invoiced_on=in_days(-1),
            due_on=dt.date.today(),
            status=Invoice.SENT,
        )
        self.client.force_login(invoice.owned_by)

        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, status=Invoice.IN_PREPARATION),
        )
        self.assertContains(
            response,
            "Moving status from &#x27;Sent&#x27; to &#x27;In preparation&#x27;."
            " Are you sure?",
        )

    def test_send_invoice_with_past_invoice_date(self):
        """Advancing the status from in preparation with a past invoice date
        emits a warning"""
        invoice = factories.InvoiceFactory.create(
            title="Test",
            subtotal=20,
            invoiced_on=in_days(-7),
            due_on=dt.date.today(),
            status=Invoice.IN_PREPARATION,
        )
        self.client.force_login(invoice.owned_by)

        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, status=Invoice.SENT),
        )
        self.assertContains(response, "with an invoice date in the past")

    def test_change_paid_invoice(self):
        """Changing paid invoices is possible too"""
        invoice = factories.InvoiceFactory.create(
            title="Test",
            subtotal=20,
            invoiced_on=in_days(-1),
            due_on=dt.date.today(),
            closed_on=dt.date.today(),
            status=Invoice.PAID,
            postal_address="Test\nStreet\nCity",
        )
        self.client.force_login(invoice.owned_by)

        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, status=Invoice.IN_PREPARATION),
        )
        self.assertContains(
            response,
            "Moving status from &#x27;Paid&#x27; to &#x27;In preparation&#x27;."
            " Are you sure?",
        )
        self.assertContains(
            response,
            "You are attempting to set status to &#x27;In preparation&#x27;,"
            " but the invoice has already been closed on {}."
            " Are you sure?".format(local_date_format(dt.date.today())),
        )

        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(
                invoice,
                status=Invoice.IN_PREPARATION,
                **{
                    WarningsForm.ignore_warnings_id: (
                        "status-unexpected status-change-but-already-closed"
                    )
                }
            ),
        )
        # print(response, response.content.decode("utf-8"))
        self.assertRedirects(response, invoice.urls["detail"])
        invoice.refresh_from_db()
        self.assertEqual(invoice.status, Invoice.IN_PREPARATION)
        self.assertIsNone(invoice.closed_on)

    def test_down_payment(self):
        """Down payment invoices can be linked to invoices"""
        project = factories.ProjectFactory.create()
        down_payment = factories.InvoiceFactory.create(
            project=project,
            type=Invoice.DOWN_PAYMENT,
            subtotal=100,
            invoiced_on=in_days(0),
            status=Invoice.SENT,
        )

        self.client.force_login(project.owned_by)
        url = project.urls["createinvoice"] + "?type=fixed"
        response = self.client.get(url)
        # print(response, response.content.decode("utf-8"))
        self.assertContains(response, down_payment.code)
        self.assertContains(response, down_payment.pretty_total_excl)
        self.assertContains(response, down_payment.pretty_status)

        response = self.client.post(
            url,
            {
                "contact": project.contact_id,
                "title": project.title,
                "description": "bla",
                "owned_by": project.owned_by_id,
                "discount": 0,
                "liable_to_vat": 1,
                "postal_address": "Anything\nStreet\nCity",
                "subtotal": 2500,
                "third_party_costs": 0,
                "apply_down_payment": [down_payment.pk],
            },
        )
        self.assertEqual(response.status_code, 302)
        invoice = project.invoices.latest("pk")
        self.assertRedirects(response, invoice.urls["detail"])

        self.assertAlmostEqual(invoice.subtotal, 2500)
        self.assertAlmostEqual(invoice.total_excl_tax, 2400)
        self.assertEqual(self.client.get(invoice.urls["pdf"]).status_code, 200)

        response = self.client.post(
            invoice.urls["update"],
            {
                "contact": project.contact_id,
                "title": project.title,
                "description": "bla",
                "owned_by": project.owned_by_id,
                "discount": 0,
                "liable_to_vat": 1,
                "postal_address": "Anything\nStreet\nCity",
                "subtotal": 2500,
                "third_party_costs": 0,
                "apply_down_payment": [],
                "status": Invoice.IN_PREPARATION,
            },
        )
        self.assertRedirects(response, invoice.urls["detail"])

        invoice.refresh_from_db()
        self.assertAlmostEqual(invoice.subtotal, 2500)
        self.assertAlmostEqual(invoice.total_excl_tax, 2500)
        self.assertEqual(invoice.down_payment_applied_to, None)

        down_payment.down_payment_applied_to = invoice
        down_payment.save(update_fields=["down_payment_applied_to"])
        invoice.save()

        response = self.client.post(invoice.urls["delete"])
        self.assertContains(response, "release-down-payments")

        response = self.client.post(
            invoice.urls["delete"],
            {WarningsForm.ignore_warnings_id: "release-down-payments"},
        )
        self.assertRedirects(response, invoice.urls["list"])

    def test_change_contact(self):
        """Selecting contacts belonging to different customers fails"""
        invoice = factories.InvoiceFactory.create(title="Test", subtotal=20)
        self.client.force_login(invoice.owned_by)

        response = self.client.post(
            invoice.urls["update"],
            invoice_to_dict(invoice, contact=factories.PersonFactory.create().pk),
        )
        # print(response.content.decode("utf-8"))
        self.assertContains(
            response,
            "The contact Vorname Nachname does not belong to The Organization Ltd.",
        )

    def test_status(self):
        """Test various results of the status badge"""
        today = dt.date.today()
        yesterday = in_days(-1)
        fmt = local_date_format(today)
        self.assertEqual(
            Invoice(status=Invoice.IN_PREPARATION).pretty_status,
            "In preparation since {}".format(fmt),
        )
        self.assertEqual(
            Invoice(status=Invoice.SENT, invoiced_on=today).pretty_status,
            "Sent on {}".format(fmt),
        )
        self.assertEqual(
            Invoice(
                status=Invoice.SENT,
                invoiced_on=yesterday,
                due_on=in_days(-5),
            ).pretty_status,
            "Sent on {} but overdue".format(local_date_format(yesterday)),
        )
        self.assertIn(
            "badge-warning",
            Invoice(
                status=Invoice.SENT,
                invoiced_on=yesterday,
                due_on=in_days(-5),
            ).status_badge,
        )
        self.assertEqual(
            Invoice(
                status=Invoice.SENT, invoiced_on=yesterday, last_reminded_on=today
            ).pretty_status,
            "Sent on {}, reminded on {}".format(local_date_format(yesterday), fmt),
        )
        self.assertEqual(
            Invoice(status=Invoice.PAID, closed_on=today).pretty_status,
            "Paid on {}".format(fmt),
        )
        self.assertEqual(Invoice(status=Invoice.CANCELED).pretty_status, "Canceled")

    def test_service_update(self):
        """Updating invoice services"""
        project = factories.ProjectFactory.create()
        invoice = factories.InvoiceFactory.create(
            project=project,
            customer=project.customer,
            contact=project.contact,
            type=Invoice.SERVICES,
        )

        service = invoice.services.create(title="Test", cost=50)

        self.client.force_login(project.owned_by)
        response = self.client.get(service.urls["update"])
        self.assertRedirects(response, invoice.urls["detail"])

        response = self.client.get(
            service.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)

        invoice.status = Invoice.SENT
        invoice.save()

        response = self.client.get(
            service.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            "Cannot modify a service bound to an invoice which is not in"
            " preparation anymore.",
        )

    def test_person_invoice_without_vat(self):
        """Test another branch in workbench/tools/pdf.py"""

        invoice = factories.InvoiceFactory.create(liable_to_vat=False)
        self.client.force_login(invoice.owned_by)
        response = self.client.get(invoice.urls["pdf"])
        self.assertEqual(response.status_code, 200)

    def test_cancellation_with_payment_notice(self):
        """Canceling invoices requires entering a payment notice"""
        invoice = factories.InvoiceFactory.create(
            invoiced_on=dt.date.today(),
            due_on=dt.date.today(),
            postal_address="Test\nStreet\nCity",
            status=Invoice.CANCELED,
        )

        msg = [
            (
                "payment_notice",
                ["Please provide a short reason for the invoice cancellation."],
            )
        ]

        with self.assertRaises(ValidationError) as cm:
            invoice.clean_fields()
        self.assertEqual(list(cm.exception), msg)

    def test_reminders(self):
        """The reminders view allows exporting dunning letters"""
        invoice = factories.InvoiceFactory.create(
            invoiced_on=in_days(-60),
            due_on=in_days(-45),
            status=Invoice.SENT,
        )
        factories.InvoiceFactory.create(
            customer=invoice.customer,
            contact=invoice.contact,
            invoiced_on=in_days(-60),
            due_on=in_days(-45),
            status=Invoice.SENT,
        )

        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/invoices/reminders/")
        self.assertContains(response, "Not reminded yet")
        # print(response, response.content.decode("utf-8"))

        response = self.client.post(
            "/invoices/dunning-letter/{}/".format(invoice.customer_id)
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

        invoice.refresh_from_db()
        self.assertEqual(invoice.last_reminded_on, dt.date.today())

        self.assertEqual(invoice.payment_reminders_sent_at(), [dt.date.today()])

        response = self.client.get("/invoices/reminders/")
        self.assertNotContains(response, "Not reminded yet")

    def test_reset_last_invoiced_on(self):
        """last_reminded_on < invoiced_on values are silently corrected/dropped"""
        invoice = factories.InvoiceFactory.create(
            invoiced_on=in_days(0), last_reminded_on=in_days(-15)
        )
        self.assertIsNone(invoice.last_reminded_on)

    def test_invoice_in_preparation_message(self):
        """More than one project invoice in preparation produces a warning"""
        project = factories.ProjectFactory.create()
        factories.InvoiceFactory.create(project=project)
        self.client.force_login(project.owned_by)

        url = project.urls["createinvoice"] + "?type=fixed"
        response = self.client.get(url)
        self.assertContains(
            response, "This project already has an invoice in preparation."
        )



import csv
import datetime as dt
import hashlib
import io
import re
from decimal import Decimal

from django.utils.dateparse import parse_date
from django.utils.encoding import force_str
from django.utils.text import slugify


def parse_zkb_csv(data):
    f = io.StringIO()
    f.write(force_str(data, encoding="utf-8", errors="ignore"))
    f.seek(0)
    dialect = csv.Sniffer().sniff(f.read(4096))
    f.seek(0)
    reader = csv.reader(f, dialect)
    next(reader)  # Skip first line
    entries = []
    while True:
        try:
            row = next(reader)
        except StopIteration:
            break
        if not row:
            continue
        try:
            day = dt.datetime.strptime(row[8], "%d.%m.%Y").date()
            amount = row[7] and Decimal(row[7])
            reference = row[4]
        except (AttributeError, IndexError, ValueError):
            continue
        if day and amount:
            details = next(reader)
            entries.append(
                {
                    "reference_number": reference,
                    "value_date": day,
                    "total": amount,
                    "payment_notice": "; ".join(
                        filter(None, (details[1], details[10], row[4]))
                    ),
                }
            )
    return entries


def postfinance_preprocess_notice(payment_notice):
    """Remove spaces from potential invoice numbers"""
    return re.sub(
        r"\b([0-9]{4}\s*-\s*[0-9]{4}\s*-\s*[0-9]{4})\b",
        lambda match: re.sub(r"\s+", "", match.group(0)),
        payment_notice,
    )


def postfinance_reference_number(payment_notice, day):
    """Either pull out the bank reference or create a hash from the notice"""
    match = re.search(r"\b([0-9]{6}[A-Z]{2}[0-9A-Z]{6,10})$", payment_notice)
    return "pf-{}".format(
        match.group(1)
        if match
        else hashlib.md5(
            slugify(payment_notice + day.isoformat()).encode("utf-8")
        ).hexdigest()
    )


def parse_postfinance_csv(data):
    f = io.StringIO()
    f.write(force_str(data, encoding="latin-1", errors="ignore"))
    f.seek(0)
    dialect = csv.Sniffer().sniff(f.read(4096))
    f.seek(0)
    reader = csv.reader(f, dialect)
    next(reader)  # Skip first line
    entries = []
    for row in reader:
        if not row:
            continue
        try:
            day = parse_date(row[4])
        except (IndexError, ValueError):
            continue
        if day is None or not row[2]:  # Only credit
            continue

        payment_notice = postfinance_preprocess_notice(row[1])
        entries.append(
            {
                "reference_number": postfinance_reference_number(payment_notice, day),
                "value_date": day,
                "total": Decimal(row[2]),
                "payment_notice": payment_notice,
            }
        )
    return entries



# Generated by Django 2.2.3 on 2019-07-19 09:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("expenses", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="expensereport",
            name="closed_on",
            field=models.DateField(blank=True, null=True, verbose_name="closed on"),
        )
    ]



import datetime as dt
from decimal import Decimal

from django.db.models import ProtectedError
from django.test import TestCase
from django.utils import timezone
from django.utils.translation import deactivate_all

from time_machine import travel

from workbench import factories
from workbench.audit.models import LoggedAction
from workbench.deals.models import Deal
from workbench.deals.reporting import accepted_deals
from workbench.templatetags.workbench import deal_group
from workbench.tools.formats import local_date_format
from workbench.tools.validation import in_days


class DealsTest(TestCase):
    def setUp(self):
        deactivate_all()

    def test_list(self):
        """Filtering smoke test"""
        deal = factories.DealFactory.create()
        self.client.force_login(deal.owned_by)

        self.assertEqual(self.client.get("/deals/").status_code, 200)
        self.assertEqual(self.client.get("/deals/?s=").status_code, 200)
        self.assertEqual(self.client.get("/deals/?s=all").status_code, 200)
        self.assertEqual(self.client.get("/deals/?s=20").status_code, 200)
        self.assertEqual(self.client.get("/deals/?s=30").status_code, 200)

        self.assertEqual(self.client.get("/deals/?s=42").status_code, 302)

    def test_detail(self):
        """Deals detail view"""
        deal = factories.DealFactory.create()
        deal.values.create(type=factories.ValueTypeFactory.create(), value=42)
        deal.save()

        self.client.force_login(deal.owned_by)
        response = self.client.get(deal.urls["detail"])
        # print(response, response.content.decode("utf-8"))
        self.assertContains(response, "42.00")

    def test_crud(self):
        """CRUD of deals incl. reopening etc."""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )
        self.client.force_login(person.primary_contact)

        response = self.client.post(
            "/deals/create/",
            {
                "customer": person.organization.id,
                "contact": person.id,
                "title": "Some deal",
                "probability": Deal.NORMAL,
                "owned_by": person.primary_contact_id,
            },
        )
        self.assertEqual(response.status_code, 302)
        deal = Deal.objects.get()
        self.assertIsNone(deal.closed_on)
        self.assertEqual(
            deal.pretty_status,
            "Open since {}".format(local_date_format(dt.date.today())),
        )

        response = self.client.post(
            deal.urls["set_status"] + "?status={}".format(Deal.DECLINED),
            {
                "closing_type": factories.ClosingTypeFactory.create(
                    represents_a_win=False
                ).pk,
            },
        )
        self.assertRedirects(response, deal.urls["detail"])

        response = self.client.get(deal.urls["update"])
        self.assertContains(response, "This deal is already closed.")

        deal.refresh_from_db()
        self.assertEqual(deal.closed_on, dt.date.today())
        self.assertEqual(
            deal.pretty_status,
            "Declined on {}".format(local_date_format(dt.date.today())),
        )

        response = self.client.post(
            deal.urls["set_status"] + "?status={}".format(Deal.OPEN),
        )
        self.assertRedirects(response, deal.urls["detail"])

        deal.refresh_from_db()
        self.assertIsNone(deal.closed_on)
        self.assertEqual(
            deal.pretty_status,
            "Open since {}".format(local_date_format(dt.date.today())),
        )

    def test_protected_m2m(self):
        """Deleting an attribute group where attributes exist already does not work"""
        deal = factories.DealFactory.create()
        group = factories.AttributeGroupFactory.create()
        value = group.attributes.create(title="Test", position=0)
        deal.attributes.add(value)

        with self.assertRaises(ProtectedError):
            group.delete()

    def test_values_and_attributes(self):
        """Test creating and updating deals with values and attributes"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        type1 = factories.ValueTypeFactory.create()
        type2 = factories.ValueTypeFactory.create(title="programming")

        group1 = factories.AttributeGroupFactory.create(title="G1")
        self.assertEqual(str(group1), "G1")
        attribute1_1 = group1.attributes.create(title="A1.1")
        group1.attributes.create(title="A1.2")
        group1.attributes.create(title="A1.3", is_archived=True)

        group2 = factories.AttributeGroupFactory.create(is_required=False)
        group2.attributes.create(title="A2.1")
        group2.attributes.create(title="A2.2")

        group3 = factories.AttributeGroupFactory.create(is_archived=True)
        group3.attributes.create(title="A3.1")

        response = self.client.get(Deal.urls["create"])
        self.assertContains(response, type1.title)
        self.assertContains(response, "A1.1")
        self.assertNotContains(response, "A1.3")
        self.assertContains(response, "A2.1")
        self.assertNotContains(response, "A3.1")

        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )
        self.client.force_login(person.primary_contact)

        response = self.client.post(
            "/deals/create/",
            {
                "customer": person.organization.id,
                "contact": person.id,
                "title": "Some deal",
                "probability": Deal.NORMAL,
                "owned_by": person.primary_contact_id,
                "value_{}".format(type1.id): 200,
                "value_{}".format(type2.id): "",
                "attribute_{}".format(group1.pk): attribute1_1.pk,
                "attribute_{}".format(group2.pk): "",
            },
        )
        self.assertEqual(response.status_code, 302)

        deal = Deal.objects.get()
        self.assertEqual(deal.value, 200)
        self.assertEqual(deal.values.count(), 1)

        # Deal has been saved once
        actions = LoggedAction.objects.for_model(deal).with_data(id=deal.id)
        self.assertEqual([action.action for action in actions], ["I"])

        da = deal.dealattribute_set.get()
        self.assertEqual(da.attribute, attribute1_1)
        self.assertEqual(
            str(da),
            "{} Some deal - {} - A1.1".format(
                deal.code, person.primary_contact.get_short_name()
            ),
        )

        response = self.client.get(deal.urls["update"])
        self.assertContains(
            response,
            '<input class="form-check-input" type="radio" name="attribute_%(g)s" value="%(a)s" class="my-2" required id="id_attribute_%(g)s_0" checked>'  # noqa
            % {"g": attribute1_1.group_id, "a": attribute1_1.id},
            html=True,
        )

    def test_set_status(self):
        """The deal status modal works"""
        deal = factories.DealFactory.create()
        self.client.force_login(deal.owned_by)

        response = self.client.get(deal.urls["set_status"] + "?status=10")
        self.assertNotContains(response, "closing_type")
        self.assertNotContains(response, "closing_notice")

        response = self.client.get(deal.urls["set_status"] + "?status=20")
        self.assertContains(response, "Award of contract")

        response = self.client.get(deal.urls["set_status"] + "?status=30")
        self.assertContains(response, "Reason for losing")

        response = self.client.get(deal.urls["set_status"] + "?status=40")
        self.assertEqual(response.status_code, 404)

        response = self.client.post(deal.urls["set_status"] + "?status=20")
        self.assertContains(response, "This field is required when closing a deal.")

    @travel("2020-02-18 12:00:00")
    def test_badge(self):
        """The deal badge contains expected informations"""
        self.assertEqual(
            factories.DealFactory.create().status_badge,
            '<span class="badge badge-info">Open since 18.02.2020</span>',
        )
        self.assertEqual(
            factories.DealFactory.create(
                status=Deal.ACCEPTED, closed_on=dt.date.today()
            ).pretty_status,
            "Accepted on 18.02.2020",
        )

    def test_xlsx(self):
        """Smoke test of the deals XLSX export"""
        deal = factories.DealFactory.create()

        group = factories.AttributeGroupFactory.create()
        attribute = group.attributes.create(title="Test", position=0)
        deal.attributes.add(attribute)

        deal.values.create(type=factories.ValueTypeFactory.create(), value=42)
        deal.save()

        self.client.force_login(deal.owned_by)
        response = self.client.get("/deals/?export=xlsx")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response["content-type"],
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        )

    def test_decision_expected_on_conditionally_required(self):
        """The "decision expected on" field is required when probability is high"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create()
        )
        self.client.force_login(person.primary_contact)

        response = self.client.post(
            "/deals/create/",
            {
                "customer": person.organization.id,
                "contact": person.id,
                "title": "Some deal",
                "probability": Deal.HIGH,
                "owned_by": person.primary_contact_id,
            },
        )
        self.assertContains(response, "This field is required when probability is high")

    def test_decision_expected_on_status(self):
        """The status of a deal depends on the "decision expected on" field"""
        today = dt.date.today()
        deal = Deal(decision_expected_on=today)
        self.assertEqual(
            deal.pretty_status,
            "Decision expected on {}".format(local_date_format(today)),
        )
        self.assertIn("badge-info", deal.status_badge)

        deal = Deal(decision_expected_on=in_days(-1))

        self.assertEqual(
            deal.pretty_status,
            "Decision expected on {}".format(local_date_format(in_days(-1))),
        )
        self.assertIn("badge-warning", deal.status_badge)

    def test_caveat_status(self):
        """The status changes after some days for inactive deals"""
        self.assertIn(
            "badge-info",
            Deal(created_at=timezone.now() - dt.timedelta(days=30)).status_badge,
        )
        self.assertIn(
            "badge-caveat",
            Deal(created_at=timezone.now() - dt.timedelta(days=360)).status_badge,
        )

    def test_update_with_archived_valuetype(self):
        """Deals using archived value types can be updated"""
        vt = factories.ValueTypeFactory.create(is_archived=True)
        deal = factories.DealFactory.create()
        deal.values.create(type=vt, value=200)
        deal.save()

        deal.refresh_from_db()
        self.assertEqual(deal.value, 200)

        self.client.force_login(deal.owned_by)
        response = self.client.get(deal.urls["update"])

        self.assertContains(response, 'name="value_{}"'.format(vt.id))
        self.assertContains(response, 'value="200.00"')

        self.assertNotContains(
            self.client.get(deal.urls["create"]), 'name="value_{}"'.format(vt.id)
        )

        # Deal has been saved twice (insert, and update with value)
        actions = LoggedAction.objects.for_model(deal).with_data(id=deal.id)
        self.assertEqual([action.action for action in actions], ["I", "U"])

        response = self.client.post(
            deal.urls["update"],
            {
                "customer": deal.customer_id,
                "contact": deal.contact_id,
                "title": deal.title,
                "probability": Deal.NORMAL,
                "owned_by": deal.owned_by_id,
                "value_{}".format(vt.id): 500,
            },
        )
        self.assertRedirects(response, deal.urls["detail"])
        deal.refresh_from_db()
        self.assertEqual(deal.value, 500)

        # Deal has been saved _one_ additional time
        actions = LoggedAction.objects.for_model(deal).with_data(id=deal.id)
        self.assertEqual([action.action for action in actions], ["I", "U", "U"])

    def test_update_with_archived_attribute(self):
        """Creating deals does not show archived attributes, but updating does"""
        group = factories.AttributeGroupFactory.create()
        group.attributes.create(title="ACTIVE")
        attribute = group.attributes.create(title="ARCHIVED", is_archived=True)

        deal = factories.DealFactory.create()

        self.client.force_login(deal.owned_by)
        response = self.client.get(deal.urls["update"])
        self.assertNotContains(response, "ARCHIVED")

        deal.attributes.add(attribute)
        response = self.client.get(deal.urls["update"])
        self.assertContains(response, "ARCHIVED")

        response = self.client.get(deal.urls["create"])
        self.assertNotContains(response, "ARCHIVED")

    def test_deal_group(self):
        """Deals are grouped according to their probability and how far off the
        decision is expected to come"""

        def idx(**kwargs):
            return deal_group(Deal(**kwargs))[0]

        self.assertEqual(idx(), 5)
        self.assertEqual(idx(probability=Deal.NORMAL), 4)
        self.assertEqual(idx(probability=Deal.HIGH), 3)
        self.assertEqual(
            idx(probability=Deal.HIGH, decision_expected_on=in_days(90)),
            3,
        )
        self.assertEqual(
            idx(probability=Deal.HIGH, decision_expected_on=in_days(30)),
            2,
        )
        self.assertEqual(
            idx(probability=Deal.HIGH, decision_expected_on=in_days(20)),
            1,
        )

    def test_deal_reporting(self):
        """Deal reporting: Accepted and declined deals, history of deals"""
        deal = factories.DealFactory.create(
            status=Deal.ACCEPTED, closed_on=dt.date.today()
        )
        vt1 = factories.ValueTypeFactory.create(position=1)
        vt2 = factories.ValueTypeFactory.create(position=0)
        deal.values.create(type=vt1, value=200)
        deal.values.create(type=vt2, value=400)
        deal.save()

        self.client.force_login(deal.owned_by)

        response = self.client.get(
            "/report/accepted-deals/?date_from=2020-01-01&date_until=2099-01-01"
        )
        self.assertContains(response, "Accepted deals")

        stats = accepted_deals([dt.date(2020, 1, 1), dt.date(2099, 1, 1)])
        self.assertEqual(
            stats["by_valuetype"],
            [
                {"type": vt2, "target": None, "sum": Decimal("400")},
                {"type": vt1, "target": None, "sum": Decimal("200")},
            ],
        )
        self.assertEqual(len(stats["by_user"]), 1)
        self.assertEqual(stats["count"], 1)
        self.assertEqual(stats["sum"], Decimal("600"))

        stats = accepted_deals([dt.date(2020, 1, 1), dt.date(2099, 1, 1)], users=[])
        self.assertEqual(stats["sum"], Decimal("0"))

        response = self.client.get("/report/declined-deals/")
        self.assertContains(response, "Declined deals")

        response = self.client.get("/report/deal-history/")
        self.assertContains(response, "Deal history")

    def test_related_offers(self):
        """Offers can be linked to deals"""
        deal = factories.DealFactory.create()
        offer = factories.OfferFactory.create(
            title="Test",
            postal_address="Test\nTest street\nTest",
            offered_on=in_days(0),
            valid_until=in_days(60),
        )

        self.client.force_login(deal.owned_by)

        # No related offers, field should not exist at all
        response = self.client.get(deal.urls["set_status"] + "?status=20")
        self.assertNotContains(response, "related_offers")

        response = self.client.post(deal.urls["add_offer"], {"modal-offer": ""})
        self.assertEqual(response.status_code, 200)

        response = self.client.post(deal.urls["add_offer"], {"modal-offer": offer.pk})
        self.assertEqual(response.status_code, 201)

        self.assertEqual(deal.related_offers.get(), offer)

        # Related offers should now appear in the form
        response = self.client.get(deal.urls["set_status"] + "?status=20")
        self.assertContains(response, "related_offers")
        self.assertContains(response, offer.code)

        # Accept the deal, and accept related offers while doing this
        closing_type = factories.ClosingTypeFactory.create(represents_a_win=True)
        response = self.client.post(
            deal.urls["set_status"] + "?status=20",
            {"closing_type": closing_type.pk, "related_offers": [offer.pk]},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        offer.refresh_from_db()
        self.assertEqual(offer.status, offer.ACCEPTED)
        self.assertTrue(offer.closed_on is not None)

        # Remove offers
        response = self.client.post(deal.urls["remove_offer"], {"modal-offer": ""})
        self.assertRedirects(response, deal.urls["detail"])

        response = self.client.post(
            deal.urls["remove_offer"], {"modal-offer": offer.pk}
        )
        self.assertRedirects(response, deal.urls["detail"])



# Generated by Django 3.1.2 on 2020-10-22 06:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("planning", "0003_delete_planningteammembership"),
    ]

    operations = [
        migrations.AddField(
            model_name="planningrequest",
            name="is_provisional",
            field=models.BooleanField(default=False, verbose_name="is provisional"),
        ),
    ]



import datetime as dt

from django.contrib import messages
from django.db import models
from django.db.models import Sum
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.expenses.rates import exchange_rates
from workbench.logbook.models import LoggedCost
from workbench.tools.formats import Z2, currency, local_date_format
from workbench.tools.models import Model, MoneyField
from workbench.tools.urls import model_urls


@model_urls
class ExpenseReport(Model):
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    created_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("created by")
    )
    closed_on = models.DateField(_("closed on"), blank=True, null=True)
    owned_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        related_name="expensereports",
        verbose_name=_("responsible"),
    )
    total = MoneyField(
        _("total"),
        default=Z2,
        blank=True,
        null=True,
        help_text=_("Total incl. tax for third-party services."),
    )

    class Meta:
        ordering = ("-created_at",)
        verbose_name = _("expense report")
        verbose_name_plural = _("expense reports")

    def __str__(self):
        return "%s, %s, %s" % (
            self.owned_by.get_full_name(),
            local_date_format(self.created_at.date()),
            currency(self.total),
        )

    def save(self, *args, **kwargs):
        if self.pk:
            self.total = self.expenses.aggregate(t=Sum("third_party_costs"))["t"] or Z2
        super().save(*args, **kwargs)

    save.alters_data = True

    def delete(self, *args, **kwargs):
        self.expenses.update(expense_report=None)
        super().delete(*args, **kwargs)

    delete.alters_data = True

    @classmethod
    def allow_create(cls, request):
        if (
            LoggedCost.objects.expenses(user=request.user)
            .filter(expense_report__isnull=True)
            .exists()
        ):
            return True
        messages.error(request, _("Could not find any expenses to reimburse."))
        return False

    @classmethod
    def allow_update(cls, instance, request):
        if instance.closed_on:
            messages.error(request, _("Cannot update a closed expense report."))
        return not instance.closed_on

    @classmethod
    def allow_delete(cls, instance, request):
        if instance.closed_on:
            messages.error(request, _("Cannot delete a closed expense report."))
        return not instance.closed_on

    @property
    def pretty_status(self):
        return (
            (_("closed on %s") % local_date_format(self.closed_on))
            if self.closed_on
            else _("In preparation")
        )

    @property
    def status_badge(self):
        return format_html(
            '<span class="badge badge-{}">{}</span>',
            "light" if self.closed_on else "info",
            self.pretty_status,
        )


class ExchangeRatesQuerySet(models.QuerySet):
    def for_day(self, day):
        try:
            return self.get(day=day)
        except self.model.DoesNotExist:
            instance, created = self.update_or_create(
                day=day, defaults={"rates": exchange_rates(day)}
            )
            return instance

    def newest(self):
        try:
            return self.latest()
        except self.model.DoesNotExist:
            return self.for_day(dt.date.today())


class ExchangeRates(models.Model):
    day = models.DateField(_("day"), unique=True)
    rates = models.JSONField()

    objects = ExchangeRatesQuerySet.as_manager()

    class Meta:
        get_latest_by = "day"
        ordering = ["-day"]
        verbose_name = _("exchange rate")
        verbose_name_plural = _("exchange rates")

    def __str__(self):
        return str(self.day)



# Generated by Django 3.0.2 on 2020-01-06 16:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0009_auto_20191210_1048"),
    ]

    operations = [
        migrations.AddField(
            model_name="absence",
            name="reason",
            field=models.CharField(
                choices=[
                    ("vacation", "vacation"),
                    ("sickness", "sickness"),
                    ("other", "other reasons"),
                ],
                default="",
                max_length=10,
                verbose_name="reason",
            ),
            preserve_default=False,
        ),
        migrations.RunSQL(
            """
            UPDATE awt_absence
            SET reason=CASE
                WHEN is_vacation THEN 'vacation'
                WHEN description ~* '.*(krank|erkältung|grippe|arzt|sickness).*' THEN 'sickness'
                ELSE 'other'
            END
            """,
            "",
        ),
    ]



import datetime as dt

from django.template import Context, Template, TemplateSyntaxError
from django.test import TestCase
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.templatetags.workbench import field_value_pairs, link_or_none


class TemplateTagsTest(TestCase):
    def test_mark_current(self):
        """{% mark_current %}"""
        t = Template(
            """
            {% load mark_current %}
            {% mark_current path %}
            <a href="/a/">a</a>
            <a href="/b/">b</a>
            <a class="c" href="/c/">c</a>
            <a class="d" href="/d/">d</a>
            <ul>
            <li><a href="/e/">e</a></li>
            <li class="f"><a href="/f/">f</a></li>
            <li class="g"><a class="g" href="/g/">g</a></li>
            </ul>
            {% endmark_current %}
            """
        )

        for path, exists, notexists in [
            (
                "/a/bla/",
                '<a class="active" href="/a/">a</a>',
                '<a class="active" href="/b/">',
            ),
            (
                "/c/bla/",
                '<a class="c active" href="/c/">c</a>',
                '<a class="d active" href="/d/">',
            ),
            ("/e/bla/", '<li class="active"><a class="active" href="/e/">', ""),
            # Small inconsistency: there is a class so only classes are changed,
            # but the anchor has no class and so gets no .active class either.
            ("/f/bla/", '<li class="f active"><a href="/f/">', ""),
            ("/g/bla/", '<li class="g active"><a class="g active" href="/g/">', ""),
        ]:
            with self.subTest(path=path, exists=exists, notexists=notexists):
                html = t.render(Context({"path": path}))
                self.assertIn(exists, html)
                if notexists:
                    self.assertNotIn(notexists, html)

    def test_invalid(self):
        """Invalid mark_current usage (no argument)"""
        with self.assertRaises(TemplateSyntaxError):
            Template("{% load mark_current %}{% mark_current %}")

    def test_bar(self):
        """{% bar %} template tag"""
        t = Template("{% load workbench %}{% bar value one %}")
        self.assertEqual(
            t.render(Context({"value": 10, "one": 100})),
            '<div class="progress progress-line" title="10%">'
            '<div class="progress-bar bg-success" role="progressbar"'
            ' style="width:10%"></div></div>',  # noqa
        )
        self.assertEqual(
            t.render(Context({"value": 80, "one": 100})),
            '<div class="progress progress-line" title="80%"><div class="progress-bar bg-success" role="progressbar" style="width:75%"></div><div class="progress-bar bg-caveat" role="progressbar" style="width:5%"></div></div>',  # noqa
        )
        self.assertEqual(
            t.render(Context({"value": 150, "one": 100})),
            '<div class="progress progress-line" title="150%"><div class="progress-bar bg-success" role="progressbar" style="width:50.0%"></div><div class="progress-bar bg-caveat" role="progressbar" style="width:16.67%"></div><div class="progress-bar bg-danger" role="progressbar" style="width:33.33%"></div></div>',  # noqa
        )

    def test_pie(self):
        """{% pie %} template tag"""
        t = Template("{% load workbench %}{% pie 1 3 %}")
        self.assertEqual(
            t.render(Context()),
            """<svg width="22" height="22" class="pie bad" style="display: inline-block">
  <circle r="10" cx="10" cy="10" class="pie-circle" />
  <path d="M 10 0 A 10 10 0 0 1 18.66025403784439 14.999999999999998 L 10 10 z" class="pie-arc" />
</svg>""",  # noqa
        )

        t = Template("{% load workbench %}{% pie 1 0 %}")
        self.assertEqual(
            t.render(Context()),
            """<svg width="22" height="22" class="pie bad" style="display: inline-block">
  <circle r="10" cx="10" cy="10" class="pie-circle" />
  <path d="M 10 0 A 10 10 0 0 1 10.0 0.0 L 10 10 z" class="pie-arc" />
</svg>""",  # noqa
        )

    def test_link_or_none(self):
        """{% link_or_none %} with special values"""
        self.assertEqual(link_or_none(0), 0)
        self.assertEqual(str(link_or_none(None)), "&ndash;")

    def test_field_value_pairs(self):
        """The |field_value_pairs filter does what it should"""
        deactivate_all()
        absence = factories.AbsenceFactory.create(starts_on=dt.date(2020, 1, 1))
        pairs = dict(field_value_pairs(absence))

        self.assertEqual(pairs["Starts on"], "01.01.2020")
        self.assertEqual(pairs["Ends on"], None)
        self.assertEqual(pairs["Reason"], "vacation")
        self.assertEqual(pairs["Is vacation"], "yes")



import json

from workbench.audit.models import LoggedAction
from workbench.tools import admin


@admin.register(LoggedAction)
class LoggedActionAdmin(admin.ModelAdmin):
    list_display = ("__str__", "data", "user_name", "created_at")
    list_filter = ["action", "table_name"]
    ordering = ("-created_at",)

    def data(self, instance):
        return json.dumps(
            instance.changed_fields if instance.action == "U" else instance.row_data
        )



# Generated by Django 2.2.5 on 2019-09-16 09:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("offers", "0006_offer_total_excl_tax")]

    operations = [
        migrations.RunSQL(
            "SET SESSION application_name TO 'Migrations';"
            "UPDATE offers_offer SET total_excl_tax=subtotal - discount;"
        )
    ]



from django.core.management.base import BaseCommand
from django.db import connections


class Command(BaseCommand):
    help = "Fixes all sequences to yield non-existing keys afterwards"

    def handle(self, **options):
        cursor = connections["default"].cursor()

        # Thanks, https://wiki.postgresql.org/wiki/Fixing_Sequences
        cursor.execute(
            """
SELECT
 'SELECT SETVAL(' ||
 quote_literal(quote_ident(PGT.schemaname) || '.' || quote_ident(S.relname)) ||
 ', COALESCE(MAX(' ||quote_ident(C.attname)|| '), 1) ) FROM ' ||
 quote_ident(PGT.schemaname)|| '.'||quote_ident(T.relname)|| ';'
FROM
 pg_class AS S,
 pg_depend AS D,
 pg_class AS T,
 pg_attribute AS C,
 pg_tables AS PGT
WHERE S.relkind = 'S'
 AND S.oid = D.objid
 AND D.refobjid = T.oid
 AND D.refobjid = C.attrelid
 AND D.refobjsubid = C.attnum
 AND T.relname = PGT.tablename
ORDER BY S.relname"""
        )

        statements = list(cursor)

        for row in statements:
            self.stdout.write(row[0])
            cursor.execute(row[0])



import io
import zipfile

from django.conf import settings
from django.utils.text import slugify
from django.utils.translation import activate, gettext as _

from workbench.credit_control.models import CreditEntry, Ledger
from workbench.invoices.models import Invoice
from workbench.tools.pdf import PDFDocument
from workbench.tools.xlsx import WorkbenchXLSXDocument


def append_invoice(*, zf, ledger_slug, invoice):
    with io.BytesIO() as buf:
        pdf = PDFDocument(buf)
        pdf.init_letter()
        pdf.process_invoice(invoice)
        pdf.generate()

        zf.writestr(
            "%s-%s/%s.pdf"
            % (ledger_slug, invoice.closed_on.strftime("%Y.%m"), invoice.code),
            buf.getvalue(),
        )


def paid_debtors_zip(date_range, *, file):
    activate(settings.WORKBENCH.PDF_LANGUAGE)
    xlsx = WorkbenchXLSXDocument()

    with zipfile.ZipFile(file, "w", compression=zipfile.ZIP_DEFLATED) as zf:
        for ledger in Ledger.objects.all():
            rows = []
            for entry in (
                CreditEntry.objects.filter(ledger=ledger, value_date__range=date_range)
                .order_by("value_date")
                .select_related("invoice__project", "invoice__owned_by")
            ):
                rows.append(
                    (
                        entry.value_date,
                        entry.total,
                        entry.payment_notice,
                        entry.invoice,
                        entry.notes,
                    )
                )

                if entry.invoice:
                    append_invoice(
                        zf=zf,
                        ledger_slug=slugify(ledger.name),
                        invoice=entry.invoice,
                    )

            xlsx.add_sheet(slugify(ledger.name))
            xlsx.table(
                (
                    _("value date"),
                    _("total"),
                    _("payment notice"),
                    _("invoice"),
                    _("notes"),
                ),
                rows,
            )

        rows = []
        for invoice in (
            Invoice.objects.filter(closed_on__range=date_range, status=Invoice.PAID)
            .exclude(
                pk__in=CreditEntry.objects.filter(invoice__isnull=False).values(
                    "invoice"
                )
            )
            .order_by("closed_on")
            .select_related("project")
        ):
            rows.append(
                (invoice.closed_on, invoice.total, invoice.payment_notice, invoice)
            )

            append_invoice(zf=zf, ledger_slug="unknown", invoice=invoice)

        xlsx.add_sheet("unknown")
        xlsx.table(
            (_("closed on"), _("total"), _("payment notice"), _("invoice")), rows
        )

        with io.BytesIO() as buf:
            xlsx.workbook.save(buf)
            zf.writestr("debtors.xlsx", buf.getvalue())



# Generated by Django 2.2.2 on 2019-06-30 20:20

from django.db import migrations


def forwards(apps, schema_editor):
    Model = apps.get_model("invoices.Invoice")
    for instance in Model.objects.select_related("customer", "contact"):
        instance._fts = " ".join(
            str(part)
            for part in [
                "%s-%04d-%04d"
                % (
                    instance.project.created_at.year,
                    instance.project._code,
                    instance._code,
                )
                if instance.project
                else "%05d" % instance._code,
                instance.customer.name,
                instance.contact.given_name if instance.contact else "",
                instance.contact.family_name if instance.contact else "",
                instance.project.title if instance.project else "",
            ]
        )
        instance.save()


class Migration(migrations.Migration):

    dependencies = [("invoices", "0012_invoice__fts")]

    operations = [migrations.RunPython(forwards)]



# Generated by Django 2.1.7 on 2019-03-07 08:25

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("invoices", "0004_recurringinvoice")]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="third_party_costs",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0.00"),
                help_text="Only used for statistical purposes.",
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="third party costs",
            ),
        ),
        migrations.AddField(
            model_name="recurringinvoice",
            name="third_party_costs",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0.00"),
                help_text="Only used for statistical purposes.",
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="third party costs",
            ),
        ),
    ]



# Generated by Django 3.0.5 on 2020-05-06 19:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("timer", "0006_delete_timerstate"),
    ]

    operations = [
        migrations.AlterField(
            model_name="timestamp",
            name="type",
            field=models.CharField(
                choices=[
                    ("start", "Start"),
                    ("stop", "Stop"),
                    ("logbook", "Logbook"),
                    ("break", "Break"),
                ],
                max_length=10,
                verbose_name="type",
            ),
        ),
        migrations.RunSQL(
            "UPDATE timer_timestamp SET type='stop' WHERE type='split';", ""
        ),
    ]



from admin_ordering.admin import OrderableAdmin

from workbench.deals import models
from workbench.tools import admin


class ValueInline(admin.TabularInline):
    model = models.Value
    extra = 0


@admin.register(models.Deal)
class DealAdmin(admin.ModelAdmin):
    inlines = [ValueInline]
    list_display = ["title", "owned_by", "value", "status", "probability", "created_at"]
    list_filter = ["status"]
    raw_id_fields = ["customer", "contact", "owned_by", "related_offers"]
    search_fields = ["title", "description"]


@admin.register(models.ValueType)
class ValueTypeAdmin(OrderableAdmin, admin.ReadWriteModelAdmin):
    list_display = ["title", "is_archived", "position", "weekly_target"]
    list_editable = ["position", "weekly_target"]
    ordering_field = "position"


class AttributeInline(OrderableAdmin, admin.TabularInline):
    model = models.Attribute
    extra = 0
    fk_name = "group"
    ordering_field = "position"


@admin.register(models.AttributeGroup)
class AttributeGroupAdmin(OrderableAdmin, admin.ReadWriteModelAdmin):
    inlines = [AttributeInline]
    list_display = ["title", "is_archived", "is_required", "position"]
    list_editable = ["position"]
    ordering_field = "position"


@admin.register(models.ClosingType)
class ClosingType(OrderableAdmin, admin.ReadWriteModelAdmin):
    list_display = ["title", "represents_a_win", "position"]
    list_editable = ["position"]
    ordering_field = "position"



import datetime as dt
from decimal import ROUND_UP, Decimal
from urllib.parse import urlencode

from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.text import capfirst
from django.utils.translation import gettext, gettext_lazy as _

from workbench.accounts.models import User
from workbench.logbook.models import Break, LoggedHours
from workbench.tools.formats import Z1, local_date_format


TIMESTAMPS_DETECT_GAP = 300  # seconds


class Slice(dict):
    @property
    def has_associated_log(self):
        return bool(self.get("logged_hours") or self.get("logged_break"))

    @property
    def elapsed_hours(self):
        if self.get("logged_hours"):
            return self["logged_hours"].hours
        elif self.get("logged_break"):
            seconds = self["logged_break"].timedelta.total_seconds()
        elif self.get("starts_at") and self.get("ends_at"):
            seconds = (self["ends_at"] - self["starts_at"]).total_seconds()
        elif not self.get("starts_at") and self.get("ends_at"):
            # STOP only; no way to prefill a nonzero elapsed hours value
            return Decimal(0)
        else:
            return None
        return Decimal(seconds / 3600).quantize(Z1, rounding=ROUND_UP)

    @property
    def hours_create_url(self):
        if self.has_associated_log:
            return None

        params = [
            ("rendered_on", self["day"].isoformat()),
            ("description", self["description"]),
            ("hours", self.elapsed_hours),
        ]
        if self.get("timestamp_id"):
            params.append(("timestamp", self["timestamp_id"]))
        else:  # No timestamp ID and no associated log -- it's a detected slice!
            params.append(("detected_ends_at", self["ends_at"]))

        # Keep zero too (see elapsed_hours above)
        query = urlencode([pair for pair in params if pair[1] or pair[1] == 0])

        return "{}?{}".format(reverse("logbook_loggedhours_create"), query)

    @property
    def break_create_url(self):
        if self.has_associated_log:
            return None

        params = [
            ("day", self["day"].isoformat()),
            ("description", self["description"]),
            ("starts_at", local_date_format(self.get("starts_at"), fmt="H:i:s")),
            ("ends_at", local_date_format(self.get("ends_at"), fmt="H:i:s")),
        ]
        if self.get("timestamp_id"):
            params.append(("timestamp", self["timestamp_id"]))
        else:  # No timestamp ID and no associated log -- it's a detected slice!
            params.append(("detected_ends_at", self["ends_at"]))

        return "{}?{}".format(
            reverse("logbook_break_create"),
            urlencode([pair for pair in params if pair[1]]),
        )


class TimestampQuerySet(models.QuerySet):
    def slices(self, user, *, day=None):
        """
        Create a list of slices from timestamps, logged hours and breaks

        The algorithm works as follows:

        1. Aggregate all timestamps into a list of ``Slice()`` objects. Slices
        and unsaved timestamps are also generated from logged hours and breaks
        without a timestamp linking them.
            a. STOP timestamps' ``created_at`` field denotes the end of a slice.
            b. START timestamps' ``created_at`` field denotes the start of a slice.
            c. Timestamps having a logged hours object use the ``hours`` field
               to either fill in the start of a slice (for STOP timestamps) or the
               end of a slice (for START timestamps).
            d. Breaks have a start and an end so those values are used directly
            e. Logged hours without a timestamp are treated as STOP timestamps,
               the ``created_at`` field of logged hours is treated as timestamp
               creation date.
        2. Merge overlapping slices or generate slices from gaps.
            a. A START timestamp opening a slice is merged with the next slice
               if the next timestamp wasn't a START and the START timestamp does
               not have an associated logbook entry.
            b. The overlap detection algorithm also takes logged hours into
               account by comparing the start of subsequent slices to the
               ``TIMESTAMPS_DETECT_GAP``.
            c. If the gap between two slices is longer than
               ``TIMESTAMPS_DETECT_GAP`` (300 seconds or 5 minutes) the algorithm
               detects a missing gap.
        3. Update the start and end of all slices now that slices have been merged
           and/or detected.
        """
        day = day or dt.date.today()
        entries = list(
            self.filter(user=user, created_at__date=day).select_related(
                "logged_hours__service", "logged_break"
            )
        )
        known_logged_hours = set(entry.logged_hours for entry in entries)
        logged_hours = user.loggedhours.filter(rendered_on=day).select_related(
            "service"
        )
        entries.extend(
            self.model(
                created_at=entry.created_at, type=self.model.LOGBOOK, logged_hours=entry
            )
            for entry in logged_hours
            if entry not in known_logged_hours
        )
        known_breaks = set(entry.logged_break for entry in entries)
        breaks = user.breaks.filter(starts_at__date=day)
        entries.extend(
            self.model(
                created_at=entry.ends_at, type=self.model.BREAK, logged_break=entry
            )
            for entry in breaks
            if entry not in known_breaks
        )
        entries = sorted(entries, key=lambda timestamp: timestamp.created_at)

        # 1. Create slices
        slices = []
        for entry in entries:
            slice = Slice(
                day=day,
                description=entry.logged_hours or entry.logged_break or entry.notes,
                logged_hours=entry.logged_hours,
                logged_break=entry.logged_break,
                timestamp_id=entry.id,
                is_start=entry.type == entry.START,
            )

            if entry.logged_break:
                slice["starts_at"] = entry.logged_break.starts_at
                slice["ends_at"] = entry.logged_break.ends_at
            elif entry.type == entry.START:
                slice["starts_at"] = entry.created_at
                if entry.logged_hours:
                    slice["ends_at"] = entry.created_at + dt.timedelta(
                        seconds=int(3600 * entry.logged_hours.hours)
                    )
            else:
                slice["ends_at"] = entry.created_at
                if entry.logged_hours:
                    slice["starts_at"] = entry.created_at - dt.timedelta(
                        seconds=int(3600 * entry.logged_hours.hours)
                    )

            slices.append(slice)

        # 2. Merge overlapping slices or generate slices from gaps
        previous = None
        result = []
        for slice in slices:
            if previous is None:
                result.append(slice)
                previous = slice
                continue

            if (
                previous["is_start"]
                and not previous.has_associated_log
                and not slice["is_start"]
            ):
                if not slice.has_associated_log:
                    previous.update(
                        {
                            "timestamp_id": slice["timestamp_id"],
                            "logged_hours": slice["logged_hours"],
                            "logged_break": slice["logged_break"],
                            "is_start": False,  # Not anymore
                            "description": "; ".join(
                                filter(
                                    None,
                                    [
                                        str(previous["description"]),
                                        str(slice["description"]),
                                    ],
                                )
                            ),
                            "ends_at": slice["ends_at"],
                        }
                    )
                    previous = slice
                    continue

                else:
                    gap = (slice["starts_at"] - previous["starts_at"]).total_seconds()
                    if gap <= TIMESTAMPS_DETECT_GAP:
                        result[-1] = slice
                        previous = slice
                        continue

            if slice.get("starts_at") and previous.get("ends_at"):
                gap = (slice["starts_at"] - previous["ends_at"]).total_seconds()
                if gap > TIMESTAMPS_DETECT_GAP:
                    result.append(
                        Slice(
                            day=day,
                            description="",
                            comment=gettext("<detected>"),
                            starts_at=previous["ends_at"],
                            ends_at=slice["starts_at"],
                            is_start=False,
                        )
                    )

            result.append(slice)
            previous = slice

        # 3. Fill in boundaries
        for previous, slice in zip(result, result[1:]):
            boundary = previous.get("ends_at") or slice.get("starts_at")
            slice["starts_at"] = previous["ends_at"] = boundary

        return result


class Timestamp(models.Model):
    START = "start"
    STOP = "stop"
    LOGBOOK = "logbook"
    BREAK = "break"

    TYPE_CHOICES = [
        (START, _("Start")),
        (STOP, _("Stop")),
        (LOGBOOK, capfirst(_("logbook"))),
        (BREAK, capfirst(_("break"))),
    ]

    user = models.ForeignKey(User, on_delete=models.CASCADE, verbose_name=_("user"))
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    type = models.CharField(_("type"), max_length=10, choices=TYPE_CHOICES)
    notes = models.CharField(_("notes"), max_length=500, blank=True)
    logged_hours = models.OneToOneField(
        LoggedHours,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("logged hours"),
    )
    logged_break = models.OneToOneField(
        Break,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        verbose_name=_("break"),
    )

    objects = TimestampQuerySet.as_manager()

    class Meta:
        ordering = ["-pk"]
        verbose_name = _("timestamp")
        verbose_name_plural = _("timestamps")

    def save(self, *args, **kwargs):
        assert self.type in {self.START, self.STOP}, "Not to be used for timestamps"
        super().save(*args, **kwargs)

    save.alters_data = True

    def __str__(self):
        return "{:>5} {}".format(self.pretty_time, self.notes)

    @property
    def pretty_time(self):
        return local_date_format(self.created_at, fmt="H:i")



# Generated by Django 2.1.7 on 2019-03-04 21:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("invoices", "0001_initial"), ("projects", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="project",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="invoices",
                to="projects.Project",
                verbose_name="project",
            ),
        )
    ]



import datetime as dt

from django import forms
from django.contrib import messages
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils.decorators import decorator_from_middleware
from django.utils.timezone import make_aware
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_GET, require_POST

from corsheaders.middleware import CorsMiddleware

from workbench.accounts.models import User
from workbench.timer.models import Timestamp
from workbench.tools.formats import Z1, hours, local_date_format
from workbench.tools.forms import Form, ModelForm
from workbench.tools.validation import filter_form


def timer(request):
    return render(
        request,
        "timer.html",
        {
            "hours": {
                "today": hours(request.user.hours["today"]),
                "week": hours(request.user.hours["week"]),
            },
        },
    )


class SignedEmailUserMixin:
    def clean(self):
        data = super().clean()
        if self.request.user.is_authenticated:
            data["user"] = self.request.user
        else:
            try:
                data["user"] = User.objects.get_by_signed_email(
                    self.request.POST.get("user") or self.request.GET.get("user")
                )
            except Exception:
                raise forms.ValidationError("Invalid user")
        return data


class TimestampForm(SignedEmailUserMixin, ModelForm):
    time = forms.TimeField(required=False)

    class Meta:
        model = Timestamp
        fields = ["type", "notes"]

    def __init__(self, data, *args, **kwargs):
        data = data.copy()
        data["type"] = "stop" if data.get("type") == "split" else data.get("type")
        super().__init__(data, *args, **kwargs)

    def clean(self):
        data = super().clean()
        if data.get("time"):
            data["created_at"] = make_aware(
                dt.datetime.combine(dt.date.today(), data["time"])
            )
        return data

    def save(self):
        instance = super().save(commit=False)
        instance.created_at = self.cleaned_data.get("created_at", instance.created_at)
        instance.user = self.cleaned_data["user"]
        instance.save()
        return instance


@csrf_exempt
@decorator_from_middleware(CorsMiddleware)
@require_POST
def create_timestamp(request):
    form = TimestampForm(request.POST, request=request)
    if not form.is_valid():
        return JsonResponse({"errors": form.errors.as_json()}, status=400)
    return JsonResponse({"success": str(form.save())}, status=201)


class SignedEmailUserForm(SignedEmailUserMixin, Form):
    pass


@decorator_from_middleware(CorsMiddleware)
@require_GET
def list_timestamps(request):
    form = SignedEmailUserForm(request.GET, request=request)
    if not form.is_valid():
        return JsonResponse({"errors": form.errors.as_json()}, status=400)

    user = form.cleaned_data["user"]
    slices = Timestamp.objects.slices(user)
    daily_hours = sum(
        (slice["logged_hours"].hours for slice in slices if slice.get("logged_hours")),
        Z1,
    )
    return JsonResponse(
        {
            "success": True,
            "user": str(user),
            "hours": daily_hours,
            "timestamps": [
                {
                    "timestamp": "{:>5} - {:>5} {:^7} {}".format(
                        local_date_format(slice.get("starts_at"), fmt="H:i") or "?  ",
                        local_date_format(slice.get("ends_at"), fmt="H:i") or "?  ",
                        "({})".format(hours(slice.elapsed_hours, plus_sign=True))
                        if slice.elapsed_hours is not None
                        else "?",
                        slice["description"] or "-",
                    ),
                    "elapsed": slice.elapsed_hours,
                    "comment": slice.get("comment", ""),
                }
                for slice in slices
            ],
        }
    )


@require_POST
def delete_timestamp(request, pk):
    timestamp = get_object_or_404(Timestamp.objects.filter(user=request.user), pk=pk)
    timestamp.delete()
    messages.success(
        request,
        _("%(class)s '%(object)s' has been deleted successfully.")
        % {"class": timestamp._meta.verbose_name, "object": timestamp},
    )
    return redirect("timestamps")


class DayForm(Form):
    day = forms.DateField(required=False)


@filter_form(DayForm)
def timestamps(request, form):
    request.user.take_a_break_warning(request=request)
    today = dt.date.today()
    day = form.cleaned_data["day"] or today
    slices = Timestamp.objects.slices(request.user, day=day)
    hours = sum(
        (slice["logged_hours"].hours for slice in slices if slice.get("logged_hours")),
        Z1,
    )
    return render(
        request,
        "timestamps.html",
        {
            "slices": slices,
            "hours": hours,
            "day": day,
            "previous": day - dt.timedelta(days=1),
            "next": day + dt.timedelta(days=1) if day < today else None,
            "url": request.build_absolute_uri(reverse("create_timestamp")),
            "is_today": day == today,
        },
    )



from collections import defaultdict

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.utils.translation import gettext as _

from workbench.invoices.models import RecurringInvoice
from workbench.tools.formats import currency, local_date_format


TEMPLATE = """\
{base_url}{recurring_url} ==>
{customer}
{invoiced_on}
{invoice}
{total}
{base_url}{invoice_url}

"""


def create_recurring_invoices_and_notify():
    by_owner = defaultdict(list)
    for ri in RecurringInvoice.objects.renewal_candidates():
        for invoice in ri.create_invoices():
            by_owner[invoice.owned_by].append((ri, invoice))

    for owner, invoices in by_owner.items():
        invoices = "\n".join(
            TEMPLATE.format(
                invoice=invoice,
                customer=invoice.contact.name_with_organization
                if invoice.contact
                else invoice.customer,
                total=currency(invoice.total),
                invoiced_on=local_date_format(invoice.invoiced_on),
                base_url=settings.WORKBENCH.URL,
                invoice_url=invoice.get_absolute_url(),
                recurring_url=ri.get_absolute_url(),
            )
            for (ri, invoice) in invoices
        )
        mail = EmailMultiAlternatives(
            _("recurring invoices"), invoices, to=[owner.email], bcc=settings.BCC
        )
        mail.send()



# Generated by Django 3.0.4 on 2020-03-24 07:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0020_recurringinvoice_create_invoice_on_day"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="service_period_from",
            field=models.DateField(
                blank=True, null=True, verbose_name="service period from"
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="service_period_until",
            field=models.DateField(
                blank=True, null=True, verbose_name="service period until"
            ),
        ),
    ]



from django.conf import settings

import requests


def exchange_rates(day=None):
    url = "https://api.exchangeratesapi.io/{}?base={}".format(
        day.isoformat() if day else "latest",
        settings.WORKBENCH.CURRENCY,
    )
    return requests.get(url, timeout=2).json()



# Generated by Django 3.0.4 on 2020-03-10 18:58

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("reporting", "0002_costcenter"),
        ("projects", "0016_auto_20190915_1249"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="cost_center",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="projects",
                to="reporting.CostCenter",
                verbose_name="cost center",
            ),
        ),
    ]



# Generated by Django 3.0 on 2019-12-10 09:48

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0008_auto_20191207_1412"),
    ]

    operations = [
        migrations.AddField(
            model_name="employment",
            name="green_hours_target",
            field=models.SmallIntegerField(
                blank=True, null=True, verbose_name="green hours target"
            ),
        ),
        migrations.AddField(
            model_name="employment",
            name="hourly_labor_costs",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="hourly labor costs",
            ),
        ),
        migrations.AlterField(
            model_name="employment",
            name="notes",
            field=models.CharField(blank=True, max_length=500, verbose_name="notes"),
        ),
    ]



from django.db.models import Prefetch
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, render
from django.utils.translation import gettext as _

from workbench.contacts.forms import OrganizationSearchForm, PersonAutocompleteForm
from workbench.contacts.models import Organization, Person
from workbench.generic import ListView
from workbench.tools.vcard import person_to_vcard, render_vcard_response


class OrganizationListView(ListView):
    model = Organization
    search_form_class = OrganizationSearchForm

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .prefetch_related(Prefetch("people", queryset=Person.objects.active()))
        )


def select(request):
    form = PersonAutocompleteForm(request.POST if request.method == "POST" else None)
    if form.is_valid():
        return JsonResponse(
            {"redirect": form.cleaned_data["person"].get_absolute_url()}, status=299
        )
    return render(
        request,
        "generic/select_object.html",
        {"form": form, "title": _("Jump to person")},
    )


def person_vcard(request, pk):
    person = get_object_or_404(Person, pk=pk)
    vcard = person_to_vcard(person).serialize()
    return render_vcard_response(request, vcard, subject=person.full_name)



from admin_ordering.admin import OrderableAdmin

from workbench.reporting import models
from workbench.tools import admin


@admin.register(models.Accruals)
class AccrualsModelAdmin(admin.ModelAdmin):
    list_display = ["cutoff_date", "accruals"]


@admin.register(models.CostCenter)
class CostCenter(OrderableAdmin, admin.ReadWriteModelAdmin):
    list_display = ["title", "position"]
    list_editable = ["position"]
    ordering_field = "position"



# Generated by Django 2.2.5 on 2019-09-13 07:19

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("logbook", "0011_merge_20190719_0941")]

    operations = [
        migrations.AlterField(
            model_name="loggedhours",
            name="created_by",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="loggedhours_set",
                to=settings.AUTH_USER_MODEL,
                verbose_name="created by",
            ),
        )
    ]
